/*
 * Copyright (C) 2009 Voltaire Ltd. All rights reserved.
 * Copyright (C) Mellanox Technologies Ltd. 2013.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <ibpm.h>
static void dump_table(ibpm_t *ibpm);

extern int backtrace (void **BUFFER, int SIZE);
extern char ** backtrace_symbols (void *const *BUFFER, int SIZE);
extern void backtrace_symbols_fd (void *const *BUFFER, int SIZE, int FD);

ibpm_t *p_ibpm;
static int clear_all_phy_cnts;

#define MAX_UCAST_LID  0xBFFF
#define VLR_ARRAY_SIZE 0xC000
#define TIMEOUTS_MUX_NUMBER 2
#define DEV_MAX_NUM_PORTS 254
static uint8_t pc[IB_PC_DATA_SZ];


uint32_t port2addr_base[NUM_SW_PORTS] = {0,
	0x101120, 0x102120, 0x103120, 0x104120, 0x105120,
	0x106120, 0x107120, 0x108120, 0x109120, 0x10A120,
	0x10B120, 0x10C120, 0x10D120, 0x10E120, 0x10F120,
	0x110120, 0x111120, 0x112120, 0x113120, 0x114120,
	0x115120, 0x116120, 0x117120, 0x118120
};

#define ARR_SIZE(a) (sizeof(a)/sizeof((a)[0]))

static const char *states[] = {
	"Unknown",
	"Down",
	"Initialize",
	"Armed",
	"Active"
};
#define MAX_STATES ARR_SIZE(states)

static const char *phy_states[] = {
	"Unknown",
	"Sleep",
	"Polling",
	"Disabled",
	"PortConfigurationTraining",
	"LinkUp",
	"LinkErrorRecovery",
	"Phy Test"
};
#define MAX_PHY_STATES ARR_SIZE(phy_states)

/* NOTE:
	You should increment NUM_HW_DEV when adding
	new device to ib_to_hw map
*/
int ib_to_hw[NUM_HW_DEV][NUM_SW_PORTS] = {
	{0x05a06, 18, 17, 16, 12, 11, 10, 6, 5, 4, 24, 23,
		22, 19, 21, 20, 13, 15, 14, 7, 9, 8, 1, 3, 2},
	{0x05a08, 1, 22, 2, 23, 3, 24, 9, 6, 8, 5, 7, 4, 10,
		13, 11, 14, 12, 15, 18, 21, 17, 20, 16, 19},
	{0x15a08, 22, 1, 23, 2, 24, 3, 9, 6, 8, 5, 7, 4, 10,
		13, 11, 14, 12, 15, 21, 18, 20, 17, 19, 16},
	{0x05a09, 7, 8, 9, 4, 5, 6, 24, 23, 22, 3, 2, 1, 10,
		11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21},
	{0x15a09, 7, 8, 9, 4, 5, 6, 3, 2, 1, 24, 23, 22, 10,
		11, 12, 13, 14, 15, 19, 20, 21, 16, 17, 18},
	{0x05a0b, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
		21, 22, 23, 24, 1, 2, 3, 4, 5, 6, 7, 8, 9},
	{0x15a0b, 10, 11, 12, 13, 14, 15, 19, 20, 21, 16, 17,
		18, 1, 2, 3, 22, 23, 24, 4, 5, 6, 7, 8, 9},
	{0x05a0c, 18, 17, 16, 12, 11, 10, 6, 5, 4, 24, 23, 22,
		19, 20, 21, 13, 14, 15, 7, 8, 9, 1, 2, 3},
	{0x05a0d, 7, 8, 9, 4, 5, 6, 24, 23, 22, 3, 2, 1, 10,
		11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21},
	{0x15a0d, 7, 8, 9, 4, 5, 6, 3, 2, 1, 24, 23, 22, 10,
		11, 12, 13, 14, 15, 19, 20, 21, 16, 17, 18},
	{0x05a2e, 18, 17, 16, 12, 11, 10, 6, 5, 4, 24, 23, 22,
		19, 21, 20, 13, 15, 14, 7, 9, 8, 1, 3, 2},
	{0x15a2e, 19, 20, 21, 12, 11, 10, 7, 8, 9, 24, 23, 22,
		16, 17, 18, 15, 14, 13, 4, 5, 6, 3, 2, 1},
	{0x05a2f, 18, 17, 16, 12, 11, 10, 6, 5, 4, 24, 23, 22,
		19, 21, 20, 13, 15, 14, 7, 9, 8, 1, 3, 2},
	{0x15a2f, 19, 20, 21, 12, 11, 10, 7, 8, 9, 24, 23, 22,
		16, 17, 18, 15, 14, 13, 4, 5, 6, 3, 2, 1},
	{0x15a30, 19, 20, 21, 12, 11, 10, 7, 8, 9, 24, 23, 22,
		16, 17, 18, 15, 14, 13, 4, 5, 6, 3, 2, 1},
	{0x15a31, 19, 20, 21, 12, 11, 10, 7, 8, 9, 24, 23, 22,
		16, 17, 18, 15, 14, 13, 4, 5, 6, 3, 2, 1},
	{0x05a32, 1, 22, 2, 23, 3, 24, 9, 6, 8, 5, 7, 4, 10,
		13, 11, 14, 12, 15, 18, 21, 17, 20, 16, 19},
	{0x15a32, 22, 1, 23, 2, 24, 3, 9, 6, 8, 5, 7, 4, 10,
		13, 11, 14, 12, 15, 21, 18, 20, 17, 19, 16},
	{0x05a33, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
		21, 22, 23, 24, 1, 2, 3, 4, 5, 6, 7, 8, 9},
	{0x15a33, 10, 11, 12, 13, 14, 15, 19, 20, 21, 16, 17,
		18, 1, 2, 3, 22, 23, 24, 4, 5, 6, 7, 8, 9},
	{0x05a34, 7, 8, 9, 4, 5, 6, 24, 23, 22, 3, 2, 1, 10,
		11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21},
	{0x15a34, 7, 8, 9, 4, 5, 6, 3, 2, 1, 24, 23, 22, 10,
		11, 12, 13, 14, 15, 19, 20, 21, 16, 17, 18},
	{0x15a37, 9, 6, 8, 5, 7, 4, 10, 13, 11, 14, 12, 15,
		21, 18, 20, 17, 19, 16, 22, 1, 23, 2, 24, 3},
	{0x15a38, 7, 8, 9, 22, 23, 24, 19, 20, 21, 10, 11,
		12, 16, 17, 18, 15, 14, 13, 4, 5, 6, 3, 2, 1},
	{0x15a42, 7, 8, 9, 22, 23, 24, 19, 20, 21, 10, 11,
		12, 16, 17, 18, 15, 14, 13, 4, 5, 6, 3, 2, 1},
	{0x15a40, 9, 8, 7, 6, 5, 4, 10, 11, 12, 13, 14, 15,
		21, 20, 19, 18, 17, 16, 22, 23, 24, 1, 2, 3},
	{0x15a45, 21, 12, 15, 11, 10, 7, 8, 6, 9, 24, 16, 17,
		18, 14, 13, 4, 5, 3, 2, 1, 19, 23, 20, 22},
	{0x1b924, 18, 15, 14, 13, 4, 5, 6, 3, 8, 24, 2, 1, 19,
		20, 12, 10, 7, 9, 23, 22, 16, 17, 21, 11},
};


static char *statestr[] = {
	"SMINFO_NOTACT",
	"SMINFO_DISCOVER",
	"SMINFO_STANDBY",
	"SMINFO_MASTER",
};

#define SM_STATE_STR(s)	(((unsigned)(s)) < (sizeof(statestr)/sizeof(char *)) ? statestr[s] : "???")

static char UFM_PROTOCOL_UDP[] = "udp";
static char UFM_PROTOCOL_HTTP[] = "http";
static char UFM_PROTOCOL_HTTPS[] = "https";

static inline const char * get_state_str(int s)
{
	if (s > 4 || s < 0)
		s = 0;
	return (states[s]);
}

static inline const char * get_phy_state_str(int s)
{
	if (s > 7 || s < 0)
		s = 0;
	return (phy_states[s]);
}

static inline int ibpm_dbg_active(log_t *p_log, int level)
{
	if (p_log->dbg_lvl >= level)
		return 1;
	return 0;
}

static void ibpm_log(log_t *p_log, int level, char *fmt, ...)
{
	va_list args;
	struct timeval tv;
	struct tm tm;
	char   buf[1024];
	FILE *stream;

	gettimeofday(&tv, NULL);
	localtime_r(&tv.tv_sec, &tm);
	tm.tm_year += 1900;

	va_start(args, fmt);
	vsnprintf(buf, sizeof(buf), fmt, args);
	va_end(args);

	stream = (p_log->file ? p_log->file : stdout);

	fprintf(stream, "%02d-%02d-%02d %02d:%02d:%02d.%06d D%d %s",
		tm.tm_mday, tm.tm_mon+1, tm.tm_year,
		tm.tm_hour, tm.tm_min, tm.tm_sec,
		(int)tv.tv_usec, level, buf);

	if (level < IBPM_INFO)
		fflush(stream);
}

/*
* DESCRIPTION
*	return diff between two timeval
*	structure in usec.
*/
static int diff_usec(struct timeval *t1, struct timeval *t2)
{
	int dt;

	dt = ((t2->tv_sec - t1->tv_sec)*1000000 + (t2->tv_usec - t1->tv_usec));

	return dt;
}

/*
* DESCRIPTION
*	return diff between two timeval
*	structure in msec.
*/
static int diff_msec(struct timeval *t1, struct timeval *t2)
{
	int dt;

	dt = (t2->tv_sec - t1->tv_sec)*1000 + (t2->tv_usec - t1->tv_usec)/1000;

	return dt;
}

/*
* DESCRIPTION
*	This function returns
*	True (1) if the device id
*	belongs to a switch which
*	supports xmitwait as part of
*	vendor specific performance
*	counters.
*/
static int check_devid(int devid)
{
	int i;

	for (i = 0; i < NUM_HW_DEV; i++)
		if (devid == (ib_to_hw[i][0] & 0xffff))
			return 1;

	return 0;
}

static int get_extended_info(ibpm_t *ibpm, int devid, unsigned int lid)
{
	ib_portid_t portid = { 0 };
	uint16_t cap_mask;

	if(ibpm->epc_support_info_size == MAX_DEVID_NUMBER){
		IBPM_LOG(&ibpm->log, IBPM_ERR, "Devid number exceeded (%d)\n",MAX_DEVID_NUMBER);
		return -1;
	}

	if (ib_portid_set(&portid, lid, 0, 0) < 0){
		IBPM_LOG(&ibpm->log, IBPM_ERR, "ib_portid_set failed for lid:%d \n",lid);
		return -1;
	}

	if (!pma_query_via(pc, &portid, 0 /* port */, ibpm->timeout, CLASS_PORT_INFO,ibpm->srcport)){
		IBPM_LOG(&ibpm->log, IBPM_ERR, "classportinfo query failed for lid:%d \n",lid);
		return -1;
	}

	memcpy(&cap_mask, pc + 2, sizeof(cap_mask));	/* CapabilityMask */
	cap_mask = ntohs(cap_mask);

	ibpm->epc_support_info[ibpm->epc_support_info_size].devid = devid;
	ibpm->epc_support_info[ibpm->epc_support_info_size].epc_supported = (cap_mask & 0x600);
	IBPM_LOG(&ibpm->log, IBPM_INFO, "get_extended_info: devid:%d(0x%x)"
			 " (by lid:%d) %s extended port counters. cap_mask:0x%x\n",
			 devid,
			 devid,
			 lid,
			 (ibpm->epc_support_info[ibpm->epc_support_info_size].epc_supported ? "support" : "do not support"),
			 cap_mask);
	ibpm->epc_support_info_size++;

	return 0;
}


/*
* DESCRIPTION
*	This function returns
*	True (1) if the device id
*	belongs to a switch which
*	supports extended port counters.
*/

static boolean_t is_dev_support_extended_pc(ibpm_t *ibpm, int devid, unsigned int lid)
{
	int i;
	boolean_t result = TRUE;

	for (i = 0; i < ibpm->epc_support_info_size; i++)
		if (devid == (ibpm->epc_support_info[i].devid)){
			result = ibpm->epc_support_info[i].epc_supported;
			break;
		}

	if(i == ibpm->epc_support_info_size){ /* no info for devid */
		if(get_extended_info(ibpm,devid,lid) < 0){
			result = FALSE;
		} else {
			result = ibpm->epc_support_info[ibpm->epc_support_info_size-1].epc_supported;
		}
	}

	IBPM_LOG(&ibpm->log, IBPM_DEBUG, "is_dev_support_extended_pc devid:0x%x(%u) : %s \n",
			 devid, devid, (result ? "TRUE" : "FALSE"));

	return result;
}


/*
* DESCRIPTION
*	This function sends the message
*	(header + message) to the socket.
*/
static void send_msg(ibpm_t *ibpm, unsigned int len)
{
	int rc;

	if (!strcmp(ibpm->ufm_protocol, UFM_PROTOCOL_HTTP) || !strcmp(ibpm->ufm_protocol, UFM_PROTOCOL_HTTPS)) {
		static int error_count = 0;
		long response;

		if (!ibpm->httphandle) {
			IBPM_LOG(&ibpm->log, IBPM_ERR, "ERR: send_msg: %s connection is not initialized.\n", ibpm->ufm_protocol);
			return;
		}

		rc = HTTPClient_post(ibpm->httphandle, ibpm->ufm_rest_url,
				(char*)ibpm->message, len+sizeof(ibpm->message->header), &response);
		if (rc != HTTPClient_OK || !(response == 200 || response == 202)) {
			if (error_count % 60 == 0)
				IBPM_LOG(&ibpm->log, IBPM_ERR, "ERR: send_msg: rc=%d, response=%d.\n", rc, response);
			error_count++;

		}
		else {
			if (error_count)
				error_count = 0;
		}
	}
	else {
		rc = sendto(ibpm->sockfd, ibpm->message,
				len+sizeof(ibpm->message->header), 0,
				(struct sockaddr *)&ibpm->servaddr, sizeof(ibpm->servaddr));
		if (rc < 0)
			IBPM_LOG(&ibpm->log, IBPM_ERR, "ERR: send_msg: sendto(%d) %m", errno);
	}
}

/*
* DESCRIPTION
*	This function gets a trid and a pointer to index.
*	It search for the slot of the trid in the trid array.
*	It sets the index to be the slot's index.
*/
static int check_tr_time(ibpm_t *ibpm, int *index)
{
	record_t *rarray = ibpm->table[ibpm->active_tbl].rarray;
	int i;
	trid_arr_t *arr;
	int ms_diff;
	struct timeval curtime;
	int ret = -1;
	int rid;
	int zero_trid = 0;

	arr = ibpm->trid_arr;
	gettimeofday(&curtime, NULL);

	/* drop all timedout entries in the trid array */
	for (i = 0; i < ibpm->window_size; i++) {
		if (arr[i].trid == 0) {
			zero_trid++;
			continue;
		}
		ms_diff = diff_msec(&arr[i].time, &curtime);
		if (ms_diff > ibpm->timeout) {
			rid = arr[i].index;
			IBPM_LOG(&ibpm->log, IBPM_ERR,
					 "ERR: dropped trid=0x%08lx (id=%d) lid=%d port=%d"
					 "type %d on timeout %d ms\n",
					 arr[i].trid, i, rarray[rid].lid,
					 rarray[rid].port, rarray[rid].type, ms_diff);
			*index = i;
			arr[i].trid = 0;
			ibpm->sent--;
			ibpm->stats.dropped++;
			ret = 0;
		}
	}

	if (zero_trid == ibpm->window_size) {
		IBPM_LOG(&ibpm->log, IBPM_ERR,
			"CRIT: clear sent(%d) since all trid are zero\n", ibpm->sent);
		ibpm->sent = 0;
	}

	return ret;
}

/*
* DESCRIPTION
*	This function gets a pointer to index.
*	It search for the first empty slot in the trid array.
*	It sets the index to be the slot's index.
*/
static int find_free_tr(ibpm_t *ibpm, int *index)
{
	trid_arr_t *arr;
	int i;

	arr = ibpm->trid_arr;

	/* check for an empty slot in the trid array */
	for (i = 0; i < ibpm->window_size; i++) {
		if (arr[i].trid == 0) {
			*index = i;
			return 1;
		}
	}

	return 0;
}

/*
* DESCRIPTION
*	This function gets a trid and a pointer to index.
*	It search for the slot of the trid in the trid array.
*	It sets the index to be the slot's index.
*/
static int find_next_tr(ibpm_t *ibpm, uint32_t trid, int *index)
{
	trid_arr_t *arr;
	int i;
	int found = 0;

	arr = ibpm->trid_arr;

	/* check for an empty slot in the trid array */
	for (i = 0; i < ibpm->window_size; i++) {
		/* search for trid match first */
		if (arr[i].trid == trid) {
			arr[i].trid = 0;
			*index = i;
			/* decrease sent MAD amount */
			ibpm->sent--;
			return 1;
		}
		/* if no match was found return available slot instead*/
		if (arr[i].trid == 0) {
			*index = i;
			found = 1;
		}
	}

	return found;
}

/*
* DESCRIPTION
*	after sending the record update
*	its time.
*/
static void update_rec_time(record_t *rec, struct timeval *curtime)
{
	int interval;

	interval = rec->isample;
	if (rec->type == MAD_TYPE_PC &&
		!rec->is_support_extended_pc &&
		p_ibpm->reset_mode != NO_RESET &&
		rec->isample > p_ibpm->default_poll_time) {
			interval = p_ibpm->default_poll_time;
			rec->isample -= p_ibpm->default_poll_time;
	}

	rec->cur_time = *curtime;
	rec->send_time.tv_sec = curtime->tv_sec + interval;
	rec->send_time.tv_usec = curtime->tv_usec;
}

/*
* DESCRIPTION
*	reset data counters and update
*	record's time.
*/
static void reset_record(record_t *rec)
{
	/* zero counters and interval */
	rec->last_time = rec->cur_time;
	rec->isample = rec->ireport;
	rec->to_report = 0;
	update_rec_time(rec, &rec->cur_time);
}

/*
* DESCRIPTION
*	This function format the xml message.
*	It prints all the non-zero error counters
*	and all of the data counters.
*/
static unsigned int format_msg(ibpm_t *ibpm, record_t *rec, unsigned int len)
{
	char *p = ibpm->message->data;
	log_t *p_log = &ibpm->log;
	unsigned int usec_interval;

	usec_interval = diff_usec(&rec->last_time, &rec->cur_time);
	if (usec_interval < rec->ireport) {
		IBPM_LOG(p_log, IBPM_ERR, "ERR: usec_interval (%u) < (%u) ireport\n",
				 usec_interval, rec->ireport);
		goto Exit;
	}
	p += len;
	p += sprintf(p, "\t\t<port lid=\"%u\" number=\"%u\" ",
				 rec->lid, rec->port);

	/* report only non-zero error counters */
	if (rec->pcount.symbolerrors != 0)
		p += sprintf(p, "SymbolErrors=\"%u\" ",
				rec->pcount.symbolerrors);
	if (rec->pcount.linkrecovers != 0)
		p += sprintf(p, "LinkRecovers=\"%u\" ",
				rec->pcount.linkrecovers);
	if (rec->pcount.linkdowned != 0)
		p += sprintf(p, "LinkDowned=\"%u\" ",
				rec->pcount.linkdowned);
	if (rec->pcount.rcverrors != 0)
		p += sprintf(p, "RcvErrors=\"%u\" ",
				rec->pcount.rcverrors);
	if (rec->pcount.rcvremotephyerrors != 0)
		p += sprintf(p, "RcvRemotePhysErrors=\"%u\" ",
				rec->pcount.rcvremotephyerrors);
	if (rec->pcount.rcvswrelayerrors != 0)
		p += sprintf(p, "RcvSwRelayErrors=\"%u\" ",
				rec->pcount.rcvswrelayerrors);
	if (rec->pcount.xmtdiscards != 0)
		p += sprintf(p, "XmtDiscards=\"%u\" ",
				rec->pcount.xmtdiscards);
	if (rec->pcount.xmtconstrainterrors != 0)
		p += sprintf(p, "XmtConstraintErrors=\"%u\" ",
				rec->pcount.xmtconstrainterrors);
	if (rec->pcount.rcvconstrainterrors != 0)
		p += sprintf(p, "RcvConstraintErrors=\"%u\" ",
				rec->pcount.rcvconstrainterrors);
	if (rec->pcount.linkintegrityerrors != 0)
		p += sprintf(p, "LinkIntegrityErrors=\"%u\" ",
				rec->pcount.linkintegrityerrors);
	if (rec->pcount.excbufoverrunerrors != 0)
		p += sprintf(p, "ExcBufOverrunErrors=\"%u\" ",
				rec->pcount.excbufoverrunerrors);
	if (rec->pcount.vl15dropped != 0)
		p += sprintf(p, "VL15Dropped=\"%u\" ",
				rec->pcount.vl15dropped);
	if (rec->pcount.xmtwait != 0)
		p += sprintf(p, "XmtWait=\"%llu\" ",
				(long long unsigned int) rec->pcount.xmtwait);

	/* report data counters */
	/* each data counter unit is 4 bytes */
	record_t *pc_rec = rec;
	if(rec->is_support_extended_pc)
	{
		tables_t *table = &ibpm->table[ibpm->active_tbl];
		pc_rec = &table->rarray[rec->ext_id];
	}

	p += sprintf(p, "XmtData=\"%llu\" RcvData=\"%llu\" "
			"XmtPkts=\"%llu\" RcvPkts=\"%llu\" />\n",
			(long long unsigned int) (pc_rec->pcount.xmtdata * 4),
			(long long unsigned int) (pc_rec->pcount.rcvdata * 4),
			(long long unsigned int) pc_rec->pcount.xmtpkts,
			(long long unsigned int) pc_rec->pcount.rcvpkts);

	len = p - ibpm->message->data;

Exit:
	return len;
}

/*
* DESCRIPTION
*	This function adds the prefix of
*	the xml message.
*/
static unsigned int add_prefix(ibpm_t *ibpm)
{
	char *p = ibpm->message->data;
	unsigned int  len;

	p += sprintf(p, "\n<report>\n\t<infiniband>\n");
	len = p - ibpm->message->data;

	return len;
}

/*
* DESCRIPTION
*	This function adds the suffix of
*	the xml message.
*/
static unsigned int add_suffix(ibpm_t *ibpm, unsigned int len)
{
	char *p = ibpm->message->data;
	p += len;

	p += sprintf(p, "\t</infiniband>\n</report>\n");
	len = p - ibpm->message->data;

	return len;
}

static void open_csv_file(log_t *p_log)
{
	time_t rawtime;
	struct tm *timeinfo;
	char format_time[30];
	char csv_filename[50];

	time(&rawtime);
	timeinfo = localtime(&rawtime);

	strftime(format_time, 30, "ibpm-%d-%m-%y-%X", timeinfo);

	strcpy(csv_filename, DEFAULT_CSV_FILENAME);
	strcat(csv_filename, format_time);
	strcat(csv_filename, ".csv");

	IBPM_LOG(p_log, IBPM_ERR, "csv file name = %s\n", csv_filename);

	p_log->csvfile = fopen(csv_filename, "w+");
	if (!p_log->csvfile)
		IBPM_LOG(p_log, IBPM_ERR,
				 "ERR: Cannot open csv file \'%s\':%m\n",
				 csv_filename);
	else {
		printf("CSV file: %s\n", csv_filename);
		IBPM_LOG(p_log, IBPM_INFO, "%s opened successfully\n",
				 csv_filename);
		fprintf(p_log->csvfile, "LID, Port, DeviceID, "
				"SymbolErrors, LinkRecovers, LinkDowned, RcvErrors, "
				"RcvRemotePhysErrors, RcvSwRelayErrors, XmtDiscards, "
				"XmtConstraintErrors, RcvConstraintErrors, "
				"LinkIntegrityErrors, ExcBufOverrunErrors, "
				"VL15Dropped, XmtWait, XmtDataKB, RcvDataKB, "
				"XmtPkts, RcvPkts, Timeouts, Status\n");
	}
}
/*
* DESCRIPTION
*	write csv file
*/
static void dump_csv_file(record_t *rec, log_t *p_log)
{
	/* if status is not ok */
	if (rec->status != 0) {
		fprintf(p_log->csvfile, "%u, %u, 0x%04x, x, x, "
			"x, x, x, x, x, x, x, x, x, x, x, "
			"x, x, x, x, %u, %u\n",
			rec->lid, rec->port, rec->devid,
			rec->pc_timeouts, rec->status);
		return;
	}

	/*
	 * xmtdata and rcvdata are in kb
	 * counters are in dword so muliplie it by 4
	 * and divide it by 1024
	 */
	fprintf(p_log->csvfile, "%u, %u, 0x%04x, %u, %u, "
			"%u, %u, %u, %u, %u, %u, %u, %u, %u, %u, %llu, "
			"%llu, %llu, %llu, %llu, %u, %u\n",
			rec->lid, rec->port, rec->devid,
			rec->pcount.symbolerrors,
			rec->pcount.linkrecovers,
			rec->pcount.linkdowned,
			rec->pcount.rcverrors,
			rec->pcount.rcvremotephyerrors,
			rec->pcount.rcvswrelayerrors,
			rec->pcount.xmtdiscards,
			rec->pcount.xmtconstrainterrors,
			rec->pcount.rcvconstrainterrors,
			rec->pcount.linkintegrityerrors,
			rec->pcount.excbufoverrunerrors,
			rec->pcount.vl15dropped,
			(long long unsigned int) rec->pcount.xmtwait,
			(long long unsigned int) rec->pcount.xmtdata / 256,
			(long long unsigned int) rec->pcount.rcvdata / 256,
			(long long unsigned int) rec->pcount.xmtpkts,
			(long long unsigned int) rec->pcount.rcvpkts,
			rec->pc_timeouts, rec->status);
}

/*
* DESCRIPTION
*	Set the message header.
*/
static void set_msg_hdr(ibpm_t *ibpm, record_t *rec)
{
	header_t *p_hdr = &ibpm->message->header;

	p_hdr->source_guid = rec->guid;
	p_hdr->sec = rec->cur_time.tv_sec;
	p_hdr->usec = rec->cur_time.tv_usec;
}

/*
* DESCRIPTION
*	This function report all of the
*	ready to report records.
*	It format the xml message and write
*	the data to csv file if required.
*/
static int ibpm_report(ibpm_t *ibpm)
{
	unsigned int len = 0;
	record_t *rec;
	int i;
	uint64_t group_guid = 0;
	log_t *p_log = &ibpm->log;
	tables_t *table = &ibpm->table[ibpm->active_tbl];
	int sindex;

	for (i = 0; i < table->rsize; i++) {
		rec = &table->rarray[i];
		/* check if xmitwait counters should be reported
		 * from vendor specific */
		/* ignore vendor record (rec->port == 0)*/
		if (rec->vendor_xmitwait && rec->port != 0 && !ibpm->errrate) {

			/* record's sw_id points to vendor record in rarray
			   and vendor record's sw_id points to switch record
			   in sarray. switch record in sarray holds all xmitwait
			   counters per switch */
			sindex = table->rarray[rec->sw_id].sw_id;
			if (sindex < 0) {
				IBPM_LOG(p_log, IBPM_INFO,
					 "invalid sindex: rec index:%d lid=%d, port=%d, devid=0x%x, sindex=%d\n",
					 i, rec->lid, rec->port, rec->devid, sindex);
				continue;

			}
			rec->pcount.xmtwait = table->sarray[sindex].xmitwait[rec->port];
			IBPM_LOG(p_log, IBPM_DEBUG, "lid=%d, port=%d, devid=0x%x, "
					 "xmitwait=%llu sindex=%d\n",
					 rec->lid, rec->port, rec->devid, rec->pcount.xmtwait,
					 sindex);
		}
		if (!rec->to_report)
			continue;

		if (ibpm->csv && !p_log->csvfile)
			open_csv_file(p_log);

		if (rec->status == 0) {
			if (group_guid == 0) {
				group_guid = rec->guid;
				set_msg_hdr(ibpm, rec);
				len = add_prefix(ibpm);
			}

			if (group_guid == rec->guid && len < MSG_DATA_LEN_THR) {
				/* gather all the records with the same guid
				   into one report */
				len = format_msg(ibpm, rec, len);
			} else {
				if (len >= MSG_DATA_LEN_THR)
					ibpm->stats.frag_msg++;

				/* send current message */
				len = add_suffix(ibpm, len);
				send_msg(ibpm, len);

		               IBPM_LOG(p_log, IBPM_INFO,
				        "HEADER: guid:0x%016llx time sec:%llu usec:%llu\n",
		                        ibpm->message->header.source_guid,
				        ibpm->message->header.sec,
		                        ibpm->message->header.usec);

				IBPM_LOG(p_log, IBPM_INFO, "%s\n",
						 ibpm->message->data);

				/* prepare next header */
				group_guid = rec->guid;
				set_msg_hdr(ibpm, rec);
				len = add_prefix(ibpm);
				len = format_msg(ibpm, rec, len);
			}

		}
		/* write to csv file */
		if (ibpm->csv)
			dump_csv_file(rec, p_log);

		rec->status = 0;
		reset_record(rec);
	}

	/* if it is the last record */
	if (group_guid != 0) {
		len = add_suffix(ibpm, len);
		send_msg(ibpm, len);

		IBPM_LOG(p_log, IBPM_INFO,
			 "HEADER: guid:0x%016llx time sec:%llu usec:%llu\n",
			 ibpm->message->header.source_guid,
			 ibpm->message->header.sec,
			 ibpm->message->header.usec);

		IBPM_LOG(p_log, IBPM_INFO, "%s\n",
				 ibpm->message->data);
	}

	if (ibpm->csv && p_log->csvfile) {
		fclose(p_log->csvfile);
		p_log->csvfile = 0;
	}


	return 0;
}

/*
* DESCRIPTION
*	This function receive a mad buffer
*	and decode all error and data counters.
*/
static void decode_perfcounters(ibpm_t *ibpm, record_t *rec, uint8_t *mad)
{
	uint32_t xmtwait;
	uint32_t xmtdata;
	uint32_t rcvdata;
	uint32_t xmtpkts;
	uint32_t rcvpkts;
	uint32_t simulate_data;
	uint32_t simulate_pkts;
	int interval;

	/* inject errors if required
	 * in this mode cur_pc_sum contains the current counters on the simulated device
	 */
	if (ibpm->errrate) {
		if (rec->is_support_extended_pc)
			interval = rec->ireport;
		else if (rec->isample > ibpm->default_poll_time)
			interval = ibpm->default_poll_time;
		else
			interval = rec->isample;

		if ((random() % 100) < ibpm->errrate) {
			rec->cur_pc_sum.symbolerrors += interval;
			rec->pcount.symbolerrors = rec->cur_pc_sum.symbolerrors;
		}
		if ((random() % 100) < ibpm->errrate) {
			rec->cur_pc_sum.rcverrors += interval;
			rec->pcount.rcverrors = rec->cur_pc_sum.rcverrors;
		}

		rec->cur_pc_sum.xmtwait += ibpm->errrate * interval;
		rec->pcount.xmtwait = rec->cur_pc_sum.xmtwait;

		if (!rec->is_support_extended_pc) {
			/* simulate data rate according to error rate */
			simulate_data = 2621440 * ibpm->errrate * interval;
			simulate_pkts = 5120 * ibpm->errrate * interval;

			rec->cur_pc_sum.xmtdata += simulate_data;
			rec->cur_pc_sum.rcvdata  += simulate_data;
			rec->cur_pc_sum.xmtpkts += simulate_pkts;
			rec->cur_pc_sum.rcvpkts  += simulate_pkts;

			rec->pcount.xmtdata = rec->cur_pc_sum.xmtdata;
			rec->pcount.rcvdata = rec->cur_pc_sum.rcvdata;
			rec->pcount.xmtpkts = rec->cur_pc_sum.xmtpkts;
			rec->pcount.rcvpkts = rec->cur_pc_sum.rcvpkts;
		}

        /*
        IBPM_LOG(p_log, IBPM_DEBUG,
				 "decode_perfcounters lid:%u port:%u errrate:%u"
				 " simulate_data:%u xmtdata count:%llu cur:%llu new:%llu \n",
				 rec->lid, rec->port, ibpm->errrate,
				 simulate_data, rec->pcount.xmtdata, rec->cur_pc_sum.xmtdata, rec->new_pc_sum.xmtdata);
                 */

		return;
	}
	/* add the proper offset to the mad */
	mad += IB_PC_DATA_OFFS;
	mad_decode_field(mad, IB_PC_PORT_SELECT_F,
					 &rec->pcount.portselect);
	mad_decode_field(mad, IB_PC_COUNTER_SELECT_F,
					 &rec->pcount.counterselect);
	mad_decode_field(mad, IB_PC_ERR_SYM_F,
					 &rec->pcount.symbolerrors);
	mad_decode_field(mad, IB_PC_LINK_RECOVERS_F,
					 &rec->pcount.linkrecovers);
	mad_decode_field(mad, IB_PC_LINK_DOWNED_F,
					 &rec->pcount.linkdowned);
	mad_decode_field(mad, IB_PC_ERR_RCV_F,
					 &rec->pcount.rcverrors);
	mad_decode_field(mad, IB_PC_ERR_PHYSRCV_F,
					 &rec->pcount.rcvremotephyerrors);
	mad_decode_field(mad, IB_PC_ERR_SWITCH_REL_F,
					 &rec->pcount.rcvswrelayerrors);
	mad_decode_field(mad, IB_PC_XMT_DISCARDS_F,
					 &rec->pcount.xmtdiscards);
	mad_decode_field(mad, IB_PC_ERR_XMTCONSTR_F,
					 &rec->pcount.xmtconstrainterrors);
	mad_decode_field(mad, IB_PC_ERR_RCVCONSTR_F,
					 &rec->pcount.rcvconstrainterrors);
	mad_decode_field(mad, IB_PC_ERR_LOCALINTEG_F,
					 &rec->pcount.linkintegrityerrors);
	mad_decode_field(mad, IB_PC_ERR_EXCESS_OVR_F,
					 &rec->pcount.excbufoverrunerrors);
	mad_decode_field(mad, IB_PC_VL15_DROPPED_F,
					 &rec->pcount.vl15dropped);

	/* do not encode xmt/rcv data/pkts if extended pc is supported */
	if(!rec->is_support_extended_pc){

		mad_decode_field(mad, IB_PC_XMT_BYTES_F,
						 &xmtdata);
		mad_decode_field(mad, IB_PC_RCV_BYTES_F,
						 &rcvdata);
		mad_decode_field(mad, IB_PC_XMT_PKTS_F,
						 &xmtpkts);
		mad_decode_field(mad, IB_PC_RCV_PKTS_F,
						 &rcvpkts);

		rec->pcount.xmtdata = xmtdata;
		rec->pcount.rcvdata = rcvdata;
		rec->pcount.xmtpkts = xmtpkts;
		rec->pcount.rcvpkts = rcvpkts;
	}

	/* do not encode xmtwait if we use vendor xmitwait */
	if (!rec->vendor_xmitwait) {
		mad_decode_field(mad, IB_PC_XMT_WAIT_F, &xmtwait);
		rec->pcount.xmtwait = xmtwait;
	}
}

/*
* DESCRIPTION
*	This function receive a mad buffer
*	and decode all extended data counters.
*/
static void decode_perfcounters_ext(ibpm_t *ibpm, record_t *rec, uint8_t *mad)
{
	uint32_t simulate_data;
	uint32_t simulate_pkts;

	/* inject errors if required */
	if (ibpm->errrate) {
		/* simulate data rate according to error rate */
		simulate_data = 2621440 * ibpm->errrate * rec->ireport;
		simulate_pkts = 5120 * ibpm->errrate * rec->ireport;

		rec->pcount.xmtdata += simulate_data;
		rec->pcount.rcvdata  += simulate_data;
		rec->pcount.xmtpkts += simulate_pkts;
		rec->pcount.rcvpkts  += simulate_pkts;

		return;
	}
	/* add the proper offset to the mad */
	mad += IB_PC_DATA_OFFS;

	/* Ignore IB_PC_EXT_PORT_SELECT_F,   IB_PC_EXT_COUNTER_SELECT_F */

	mad_decode_field(mad, IB_PC_EXT_XMT_BYTES_F,
					 &rec->pcount.xmtdata);
	mad_decode_field(mad, IB_PC_EXT_RCV_BYTES_F,
					 &rec->pcount.rcvdata);
	mad_decode_field(mad, IB_PC_EXT_XMT_PKTS_F,
					 &rec->pcount.xmtpkts);
	mad_decode_field(mad, IB_PC_EXT_RCV_PKTS_F,
					 &rec->pcount.rcvpkts);

	/* Ignore
	IB_PC_EXT_XMT_UPKTS_F, IB_PC_EXT_RCV_UPKTS_F,
	IB_PC_EXT_XMT_MPKTS_F, IB_PC_EXT_RCV_MPKTS_F
	Not implemented by switches
	*/
}

/*
* DESCRIPTION
*	This function receive a mad buffer
*	and decode all error and data counters.
*/
static int
set_xmitwait_data(ibpm_t *ibpm, int start, int end, int id)
{
	int port;
	uint32_t base_addr;
	is3_config_space_t *cs;
	int i = 0;

	memset(ibpm->data, 0, sizeof(ibpm->data));

	cs = (is3_config_space_t *)&ibpm->data;
	for (port = start; port <= end; port++) {
		/* get the port base address from table */
		base_addr = port2addr_base[ib_to_hw[id][port]];
		/* add xmit wait offset (0x1c) to the port base address */
		cs->record[i].address = htonl(base_addr + 0x1c);
		i++;
	}
	return 0;
}

/*
* DESCRIPTION
*	This function sends the MAD.
*	If the a result is expected,
*	it update the transaction id (trid)
*	in the trid array.
*/

static int snd_mad(ibpm_t *ibpm, record_t *rec, int rec_id,
		uint8_t *umad, int agent, int len, int timeout)
{
	uint32_t trid;
	uint32_t attrid;
	int method;
	int mgmtclass;
	uint8_t *mad = umad_get_mad(umad);
	log_t *p_log = &ibpm->log;

	trid = mad_get_field64(mad, 0, IB_MAD_TRID_F);
	/* check valid trid */
	if (!trid) {
		IBPM_LOG(p_log, IBPM_ERR,
				 "CRIT: error trid=0\n");
		return -1;
	}
	attrid = mad_get_field(mad, 0, IB_MAD_ATTRID_F);
	method = mad_get_field(mad, 0, IB_MAD_METHOD_F);
	mgmtclass = mad_get_field(mad, 0, IB_MAD_MGMTCLASS_F);

	IBPM_LOG(p_log, IBPM_DEBUG,
			 "rarray[%d]: trid=0x%08lx index=%d lid=%d "
			 "port=%d class=0x%x method=0x%x attrid=0x%x\n",
			 rec_id, trid, rec->index, rec->lid, rec->port,
			 mgmtclass, method, attrid);

	/* for resetting the counter keep trid on record
	 * to know if we reset was successfully or not */
	if (method == IB_MAD_METHOD_SET &&
	    (attrid == IB_GSI_PORT_COUNTERS ||
	     attrid == IB_MLX_IS3_CONFIG_SPACE_ACCESS)) {
		rec->reset_trid[rec->type] = trid;
		IBPM_LOG(p_log, IBPM_DEBUG, "sending reset mad: "
				"trid=0x%08lx index=%d type=%d attrid=%d\n",
				trid, rec->index, rec->type, attrid);
	}

	if (timeout) {
		/* response expected */
		/* init cell in the array */
		gettimeofday(&ibpm->trid_arr[rec_id].time, NULL);
		if (ibpm->trid_arr[rec_id].trid)
			IBPM_LOG(p_log, IBPM_ERR, "CRIT: trid[%d] is not zero (0x%08x)\n",
					 rec_id, ibpm->trid_arr[rec_id].trid);
		ibpm->trid_arr[rec_id].trid = trid;
		ibpm->trid_arr[rec_id].index = rec->index;
	}

	/* send MAD */
	if (umad_send(mad_rpc_portid(ibpm->srcport), agent,
			umad, len, timeout, 0) < 0) {
		IBPM_LOG(p_log, IBPM_ERR,
				 "ERR: umad_send method set failed %m\n");
		return -1;
	}

	if (timeout) {
		ibpm->stats.mad_sent++;
		ibpm->sent++;
	}

	return 0;
}

static int setMadSl(ibpm_t *ibpm, ib_portid_t *p_portid)
{
	if (ibpm->vlr_config.lid_to_sl) {
		if (ibpm->vlr_config.lid_to_sl[p_portid->lid] == 0xFF) {
			IBPM_LOG(&ibpm->log, IBPM_ERR, "Can not find SL for lid:%u. Will not send MAD.\n",
					 p_portid->lid);
			return -1;
		}
		p_portid->sl = ibpm->vlr_config.lid_to_sl[p_portid->lid];
	}
	return 0;
}


/*
* DESCRIPTION
*	This function sends the
*	vendor specific MAD.
*	It updates the rpc header's
*	fields, build the packet and send
*	it.
*/
static int vendor_mad(ibpm_t *ibpm, int lid, uint attrid,
		uint mod, int method, record_t *rec, int rec_id)
{
	ib_rpc_t rpc = {0};
	ib_portid_t portid = {0};
	uint8_t *umad;
	int agent;
	int len;

	umad = ibpm->umad;
	memset(umad, 0, umad_size() + IB_MAD_SIZE);
	agent = mad_rpc_class_agent(ibpm->srcport, IB_MLX_VENDOR_CLASS);

	/* build MAD header */
	rpc.mgtclass = IB_MLX_VENDOR_CLASS;
	rpc.method = method;
	rpc.attr.id = attrid;
	rpc.timeout  = ibpm->timeout;
	rpc.attr.mod = mod;
	rpc.datasz = IB_VENDOR_RANGE1_DATA_SIZE;
	rpc.dataoffs = IB_VENDOR_RANGE1_DATA_OFFS;

	portid.lid = lid;
	portid.qp = 1;
	portid.qkey = IB_DEFAULT_QP1_QKEY;
	if (ibpm->vlr_config.lid_to_sl && setMadSl(ibpm, &portid))
		return -1;

	len = mad_build_pkt(umad, &rpc, &portid, NULL, &ibpm->data);

	return snd_mad(ibpm, rec, rec_id, umad, agent, len, ibpm->timeout);
}

/*
* DESCRIPTION
*	This function returns the
*	index of the current switch
*	ports mapping in the ib_to_hw matrix
*	according to the device id and revision.
*/
static int get_hw_ports(int devid, int rev)
{
	int key;
	int i;
	/*
		Concatenate the device id to the revision and check
		return the correct ib_to_hw matrix line number.
	*/
	key = rev << 16 | devid;
	for (i = 0; i < NUM_HW_DEV; i++)
		if (key == (ib_to_hw[i][0]))
			return i;

	return -1;
}

/*
* DESCRIPTION
*	This function encode and send
*	xmitwait MAD (vendor specific).
*/
int encode_xmitwait(ibpm_t *ibpm, record_t *rec, int rec_id)
{
	int lid;
	int type;
	log_t *p_log = &ibpm->log;
	tables_t *table = &ibpm->table[ibpm->active_tbl];
	int sindex  = rec->sw_id;
	sw_rec_t *p_sw = &table->sarray[sindex];
	int id;

	lid = rec->lid;
	type = rec->type;

	IBPM_LOG(p_log, IBPM_DEBUG,
			 "lid=%d, type=%d\n", lid, type);

	/* check sw_id size */
	if (sindex > table->ssize) {
		IBPM_LOG(p_log, IBPM_ERR, "CRIT: rarrary[%d] sindex=%d "
				"exceeds sarry size=%d\n",
				rec->index, sindex, table->ssize);
		return -1;
	}
	id = get_hw_ports(p_sw->devid, p_sw->rev);
	if (id < 0) {
		IBPM_LOG(p_log, IBPM_WARN, "id=%d cannot find devid=0x%x "
				"rev=%d (sindex=%d mad type=%d)\n",
				 id, p_sw->devid, p_sw->rev, rec->sw_id, rec->type);
		return -1;
	}

	/* encode MAD */
	switch (type) {
	case MAD_TYPE_LOW_XMITWAIT:
		if (set_xmitwait_data(ibpm, 1, 12, id) < 0)
			return -1;
		break;
	case MAD_TYPE_HI_XMITWAIT:
		if (set_xmitwait_data(ibpm, 13, 24, id) < 0)
			return -1;
	}

	/* send MAD */
	return vendor_mad(ibpm, lid, IB_MLX_IS3_CONFIG_SPACE_ACCESS,
			IB_MLX_CONFIG_SPACE_ATTR_MOD, IB_MAD_METHOD_GET, rec, rec_id);
}

/*
* DESCRIPTION
*	This function sends pm MAD.
*/

static int pm_mad(ibpm_t *ibpm, ib_portid_t *portid,
		int method, int attr, int timeout, uint8_t *data,
		record_t *rec, int rec_id)
{
	uint8_t *umad;
	int len = IB_MAD_SIZE;
	ib_rpc_t rpc = {0};
	int agent;

	agent = mad_rpc_class_agent(ibpm->srcport, IB_PERFORMANCE_CLASS);
	umad = ibpm->umad;
	memset(umad, 0, UMAD_SIZE);

	/* build MAD header */
	rpc.mgtclass = IB_PERFORMANCE_CLASS;
	rpc.method = method;
	rpc.attr.id = attr;

	/* Same for attribute IDs */
	rpc.attr.mod = 0;
	rpc.datasz = IB_PC_DATA_SZ;
	rpc.dataoffs = IB_PC_DATA_OFFS;

	/* build MAD */
	len = mad_build_pkt(umad, &rpc, portid, NULL, data);

	/* send MAD */
	return snd_mad(ibpm, rec, rec_id, umad, agent, len, timeout);
}

/*
* DESCRIPTION
*	This function sends a performance
*	counters MAD (pm).
*	The method is get.
*/
static int pc_mad_get(ibpm_t *ibpm, record_t *rec, int rec_id, boolean_t extended)
{
	ib_portid_t portid = {0};
	uint8_t data[IB_PC_DATA_SZ] = {0};

	portid.lid = rec->lid;
	portid.qp  = 1;
	portid.qkey = IB_DEFAULT_QP1_QKEY;
	if (ibpm->vlr_config.lid_to_sl && setMadSl(ibpm, &portid))
		return -1;

	mad_set_field(data, 0, IB_PC_PORT_SELECT_F, rec->port);

	if(extended)
		return pm_mad(ibpm, &portid, IB_MAD_METHOD_GET, IB_GSI_PORT_COUNTERS_EXT,
				ibpm->timeout, data, rec, rec_id);
	else
		return pm_mad(ibpm, &portid, IB_MAD_METHOD_GET, IB_GSI_PORT_COUNTERS,
				ibpm->timeout, data, rec, rec_id);
}

/*
* DESCRIPTION
*	This function sends a performance
*	counters MAD (pm).
*	The method is set.
*/
static int pc_mad_set(ibpm_t *ibpm, record_t *rec, int mask, int rec_id)
{
	ib_portid_t portid = {0};
	uint8_t data[IB_PC_DATA_SZ] = {0};

	portid.lid = rec->lid;
	portid.qp  = 1;
	portid.qkey = IB_DEFAULT_QP1_QKEY;
	if (ibpm->vlr_config.lid_to_sl && setMadSl(ibpm, &portid))
		return -1;

	mad_set_field(data, 0, IB_PC_PORT_SELECT_F, rec->port);

	/* set counter mask*/
	mad_set_field(data, 0, IB_PC_COUNTER_SELECT_F, mask);

	/* set xmit_wait counter mask*/
	mask = mask >> 16;
	mad_set_field(data, 0, IB_PC_COUNTER_SELECT2_F, mask);

	return pm_mad(ibpm, &portid, IB_MAD_METHOD_SET, IB_GSI_PORT_COUNTERS,
			ibpm->timeout, data, rec, rec_id);
}

/*
* DESCRIPTION
*	This function reset the data
*	counters and error counter if
*	needed.
*/
static void reset_counters(ibpm_t *ibpm, record_t *rec, int *rec_id)
{
	int bitmask = 0;
	log_t *p_log = &ibpm->log;
	struct perf_count pc_thr = ibpm->pc_thr;

	/* if counters are simulated cur_pc_sum will contain the
	 * simulated counters on the simulated device.
	 */
	if (!ibpm->errrate)
		rec->new_pc_sum = rec->cur_pc_sum;

	/* check if a counter has reached its threshold.
	 * if threshold was reached - set 1
	 * in the proper bit of the mask.
	 * also add the pcount value to the new sum.
	 * */
	if (rec->pcount.symbolerrors >= pc_thr.symbolerrors) {
		bitmask |= 0x1;
		rec->new_pc_sum.symbolerrors += rec->pcount.symbolerrors;
		if (ibpm->errrate)
			rec->cur_pc_sum.symbolerrors = 0;
	}
	if (rec->pcount.linkrecovers >= pc_thr.linkrecovers) {
		bitmask |= 0x2;
		rec->new_pc_sum.linkrecovers += rec->pcount.linkrecovers;
	}
	if (rec->pcount.linkdowned >= pc_thr.linkdowned) {
		bitmask |= 0x4;
		rec->new_pc_sum.linkdowned += rec->pcount.linkdowned;
	}
	if (rec->pcount.rcverrors >= pc_thr.rcverrors) {
		bitmask |= 0x8;
		rec->new_pc_sum.rcverrors += rec->pcount.rcverrors;
		if (ibpm->errrate)
			rec->cur_pc_sum.rcverrors = 0;
	}
	if (rec->pcount.rcvremotephyerrors >= pc_thr.rcvremotephyerrors) {
		bitmask |= 0x10;
		rec->new_pc_sum.rcvremotephyerrors += rec->pcount.rcvremotephyerrors;
	}
	if (rec->pcount.rcvswrelayerrors >= pc_thr.rcvswrelayerrors) {
		bitmask |= 0x20;
		rec->new_pc_sum.rcvswrelayerrors += rec->pcount.rcvswrelayerrors;
	}
	if (rec->pcount.xmtdiscards >= pc_thr.xmtdiscards) {
		bitmask |= 0x40;
		rec->new_pc_sum.xmtdiscards += rec->pcount.xmtdiscards;
	}
	if (rec->pcount.xmtconstrainterrors >= pc_thr.xmtconstrainterrors) {
		bitmask |= 0x80;
		rec->new_pc_sum.xmtconstrainterrors += rec->pcount.xmtconstrainterrors;
	}
	if (rec->pcount.rcvconstrainterrors >= pc_thr.rcvconstrainterrors) {
		bitmask |= 0x100;
		rec->new_pc_sum.rcvconstrainterrors += rec->pcount.rcvconstrainterrors;
	}
	if (rec->pcount.linkintegrityerrors >= pc_thr.linkintegrityerrors) {
		bitmask |= 0x200;
		rec->new_pc_sum.linkintegrityerrors += rec->pcount.linkintegrityerrors;
	}
	if (rec->pcount.excbufoverrunerrors >= pc_thr.excbufoverrunerrors) {
		bitmask |= 0x400;
		rec->new_pc_sum.excbufoverrunerrors += rec->pcount.excbufoverrunerrors;
	}
	if (rec->pcount.vl15dropped >= pc_thr.vl15dropped) {
		bitmask |= 0x800;
		rec->new_pc_sum.vl15dropped += rec->pcount.vl15dropped;
	}
	if(!rec->is_support_extended_pc){
		if (rec->pcount.xmtdata >= pc_thr.xmtdata) {
			bitmask |= 0x1000;
			rec->new_pc_sum.xmtdata += rec->pcount.xmtdata;
			if (ibpm->errrate)
				rec->cur_pc_sum.xmtdata = 0;
		}
		if (rec->pcount.rcvdata >= pc_thr.rcvdata) {
			bitmask |= 0x2000;
			rec->new_pc_sum.rcvdata += rec->pcount.rcvdata;
			if (ibpm->errrate)
				rec->cur_pc_sum.rcvdata = 0;
		}
		if (rec->pcount.xmtpkts >= pc_thr.xmtpkts) {
			bitmask |= 0x4000;
			rec->new_pc_sum.xmtpkts += rec->pcount.xmtpkts;
			if (ibpm->errrate)
				rec->cur_pc_sum.xmtpkts = 0;
		}
		if (rec->pcount.rcvpkts >= pc_thr.rcvpkts) {
			bitmask |= 0x8000;
			rec->new_pc_sum.rcvpkts += rec->pcount.rcvpkts;
			if (ibpm->errrate)
				rec->cur_pc_sum.rcvpkts = 0;
		}
	}
	/* if record support vendor xmitwait,
	 * do not reset and sum its xmitwait counter.
	 * all logic should be done via send/receive xmitwait.
	 */
	if (!rec->vendor_xmitwait || ibpm->errrate) {
		if (rec->pcount.xmtwait >= pc_thr.xmtwait) {
			bitmask |= 0x10000;
			rec->new_pc_sum.xmtwait += rec->pcount.xmtwait;
			if (ibpm->errrate)
				rec->cur_pc_sum.xmtwait = 0;
		}
		rec->pcount.xmtwait += rec->cur_pc_sum.xmtwait; //will be recalculated if ibpm->errrate
	}

	/* update pcount to hold pcount + cur_pc_sum.
	 * then, when reporting pcount it will report the
	 * actual counters (pcount + cur_pc_sum).
	 */
	if (!ibpm->errrate) {
		rec->pcount.symbolerrors += rec->cur_pc_sum.symbolerrors;
		rec->pcount.linkrecovers += rec->cur_pc_sum.linkrecovers;
		rec->pcount.linkdowned += rec->cur_pc_sum.linkdowned;
		rec->pcount.rcverrors += rec->cur_pc_sum.rcverrors;
		rec->pcount.rcvremotephyerrors += rec->cur_pc_sum.rcvremotephyerrors;
		rec->pcount.rcvswrelayerrors += rec->cur_pc_sum.rcvswrelayerrors;
		rec->pcount.xmtdiscards += rec->cur_pc_sum.xmtdiscards;
		rec->pcount.xmtconstrainterrors += rec->cur_pc_sum.xmtconstrainterrors;
		rec->pcount.rcvconstrainterrors += rec->cur_pc_sum.rcvconstrainterrors;
		rec->pcount.linkintegrityerrors += rec->cur_pc_sum.linkintegrityerrors;
		rec->pcount.excbufoverrunerrors += rec->cur_pc_sum.excbufoverrunerrors;
		rec->pcount.vl15dropped += rec->cur_pc_sum.vl15dropped;
		if(!rec->is_support_extended_pc){
			rec->pcount.xmtdata += rec->cur_pc_sum.xmtdata;
			rec->pcount.rcvdata += rec->cur_pc_sum.rcvdata;
			rec->pcount.xmtpkts += rec->cur_pc_sum.xmtpkts;
			rec->pcount.rcvpkts += rec->cur_pc_sum.rcvpkts;
		}
	} else {
        /* in simulation mode new_pc_sum holds the total counter while cur_pc_sum.
         * holds the current counter on the device
         * pcount = cur_pc_sum + new_pc_sum.
         */

		rec->pcount.symbolerrors = rec->cur_pc_sum.symbolerrors + rec->new_pc_sum.symbolerrors;
		rec->pcount.rcverrors = rec->cur_pc_sum.rcverrors + rec->new_pc_sum.rcverrors;
		rec->pcount.xmtwait = rec->cur_pc_sum.xmtwait + rec->new_pc_sum.xmtwait;

		if(!rec->is_support_extended_pc){
			rec->pcount.xmtdata = rec->cur_pc_sum.xmtdata + rec->new_pc_sum.xmtdata;
			rec->pcount.rcvdata = rec->cur_pc_sum.rcvdata + rec->new_pc_sum.rcvdata;
			rec->pcount.xmtpkts = rec->cur_pc_sum.xmtpkts + rec->new_pc_sum.xmtpkts;
			rec->pcount.rcvpkts = rec->cur_pc_sum.rcvpkts + rec->new_pc_sum.rcvpkts;
		}
	}

	if (bitmask == 0)
		return;

	if ((bitmask & 0x10fff))
		ibpm->stats.reset_error_pc++;

	IBPM_LOG(p_log, IBPM_DEBUG, "lid=%d port=%d bitmask = 0x%x\n",
			rec->lid, rec->port, bitmask);

	if (ibpm->errrate) {
		IBPM_LOG(p_log, IBPM_DEBUG, "Simulated counters skip reset counters.\n",
				rec->lid, rec->port, bitmask);
		return;
	}

	/* send reset mad */
	pc_mad_set(ibpm, rec, bitmask, *rec_id);

	/* put -1 in the rec_id (next_id)
	 * in order to know that the rec_id slot
	 * was used by the reset mad */
	*rec_id = -1;
}

/*
* DESCRIPTION
*	This function decodes the
*	vendor specific xmitwait MAD.
*/
static void decode_xmitwait(ibpm_t *ibpm, record_t *rec, uint8_t *mad,
		int start_port, int *rec_id)
{
	is3_config_space_t *cs = 0;
	int i, port;
	int reset = 0;
	log_t *p_log = &ibpm->log;
	tables_t *table = &ibpm->table[ibpm->active_tbl];
	int sindex = rec->sw_id;
	sw_rec_t *p_sw = &table->sarray[sindex];
	uint32_t val;
	uint32_t threshold = (uint32_t) ibpm->pc_thr.xmtwait;

	rec->to_report = 0;
	cs = (is3_config_space_t *) (mad + IB_VENDOR_RANGE1_DATA_OFFS);
	IBPM_LOG(p_log, IBPM_DEBUG, "lid=%d port=%d decoding xmitwait\n",
			rec->lid, rec->port);
	for (i = 0, port = start_port; i < 12; i++, port++) {
		val = ntohl(cs->record[i].data);
		if (ibpm->verbose)
			IBPM_LOG(p_log, IBPM_DEBUG, "Port %d: PortXmitWait 0x%x\n",
					port, val);
		p_sw->xmitwait_new_sum[port] = p_sw->xmitwait_cur_sum[port];
		p_sw->xmitwait[port] = val;
		if (val >= threshold) {
			reset = 1;
			cs->record[i].mask = 0xffffffff;
			cs->record[i].data = 0;
			p_sw->xmitwait_new_sum[port] += val;
			IBPM_LOG(p_log, IBPM_INFO, "reset xmit_wait lid %d port %d "
					"val=%u threshold=%u\n",
					 p_sw->lid, port, val, threshold);
		}
		p_sw->xmitwait[port] += p_sw->xmitwait_cur_sum[port];
	}

	/* reset xmitwait counters if necessary */
	if (reset) {
		/* copy cs to ibpm->data to reset all necessary ports */
		memcpy(&ibpm->data, cs, sizeof(is3_config_space_t));
		vendor_mad(ibpm, rec->lid, IB_MLX_IS3_CONFIG_SPACE_ACCESS,
				IB_MLX_CONFIG_SPACE_ATTR_MOD, IB_MAD_METHOD_SET, rec, *rec_id);
		*rec_id = -1;
	}
}

/*
* DESCRIPTION
*	This function decodes a
*	given MAD by its type.
*/
static void decode_mad(ibpm_t *ibpm, uint8_t *mad, int *id, uint32_t trid)
{
	int rindex;
	int sindex;
	int start_port;
	record_t *rec;
	tables_t *table = &ibpm->table[ibpm->active_tbl];
	log_t *p_log = &ibpm->log;
	void *dst;
	void *src;
	int len;

	rindex = ibpm->trid_arr[*id].index;
	rec = &table->rarray[rindex];

	/* check MAD type */
	switch (rec->type) {
	case MAD_TYPE_PC:
		/* if it is reset mad - copy the new sum into the current sum */
		if (trid == rec->reset_trid[MAD_TYPE_PC]) {
			dst = &rec->cur_pc_sum;
			src = &rec->new_pc_sum;
			len = sizeof(rec->cur_pc_sum);
			goto UpdateSum;
		}
		decode_perfcounters(ibpm, rec, mad);

		/* do not call reset_counters() in NO_RESET mode */
		if (ibpm->reset_mode != NO_RESET)
			reset_counters(ibpm, rec, id);
		if (rec->firstrun) {
			rec->firstrun = 0;
			reset_record(rec);
			break;
		}
		/* report when current interval has reach the report interval */
		if (diff_usec(&rec->last_time, &rec->cur_time)/1000000 >= rec->ireport)
			rec->to_report = 1;
		break;
	case MAD_TYPE_PC_EXT:
		decode_perfcounters_ext(ibpm, rec, mad);

		if (rec->firstrun) {
			rec->firstrun = 0;
			reset_record(rec);
			break;
		}
		break;

	case MAD_TYPE_LOW_XMITWAIT:
	case MAD_TYPE_HI_XMITWAIT:
		if (rec->type == MAD_TYPE_LOW_XMITWAIT)
			start_port = 1;
		else
			start_port = 13;

		/* if it is reset mad - copy the new sum into the current sum */
		if (trid == rec->reset_trid[rec->type]) {
			sindex = rec->sw_id;
			dst = &table->sarray[sindex].xmitwait_cur_sum[start_port];
			src = &table->sarray[sindex].xmitwait_new_sum[start_port];
			len = 12 * sizeof(uint64_t);
			goto UpdateSum;
		}
		decode_xmitwait(ibpm, rec, mad, start_port, id);
		break;
	}
	return;

UpdateSum:
	memcpy(dst, src, len);
	IBPM_LOG(p_log, IBPM_DEBUG,
			"got reset mad (trid=0x%08lx)) update type %d sum\n",
			trid, rec->type);
	rec->reset_trid[rec->type] = 0;
}

/*
* DESCRIPTION
*	This function receives a MAD -
*	return -1 if MAD was not received correctly
*	or 0 otherwise.
*/
static int rcv_mad(ibpm_t *ibpm, int *free_id)
{
	uint8_t *umad, *mad;
	int len = IB_MAD_SIZE;
	log_t *p_log = &ibpm->log;
	uint32_t trid;
	int rid;
	int ret = 0;
	int timeout = ibpm->timeout + 10;
	tables_t *table = &ibpm->table[ibpm->active_tbl];
	record_t *p_rec;

	/* use longer recv timeout than send timeout,
	   else can race with kernel timeouts. */
	umad = ibpm->umad;
	memset(umad, 0, UMAD_SIZE);

	/* receive MAD */
	ret = umad_recv(mad_rpc_portid(ibpm->srcport), umad, &len, timeout);
	/* check if receive succeeded */
	if (ret < 0) {
		IBPM_LOG(p_log, IBPM_ERR,
				 "ERR: sent %d timeout %d, umad_recv returned %d: %m\n",
					ibpm->sent, timeout, ret);
		ibpm->stats.socket_timeouts++;
		ret = check_tr_time(ibpm, free_id);
		return ret;
	}

	mad = umad_get_mad(umad);
	trid = mad_get_field64(mad, 0, IB_MAD_TRID_F);
	/* check valid trid */
	if (!trid) {
		IBPM_LOG(p_log, IBPM_ERR,
				 "CRIT: lid=%d port=%d error trid=0\n");
		return -1;
	}

	ret = find_next_tr(ibpm, trid, free_id);
	/* ret == 0 trid was not found in the trid array */
	/* ret == 1 trid was found */
	if (!ret) {
		IBPM_LOG(p_log, IBPM_ERR, "ERR: cannot find trid=0x%08lx"
				"in trid array\n", trid);
		return -1;
	}

	/* check received MAD status */
	if (umad_status(umad)) {
		rid = ibpm->trid_arr[*free_id].index;
		p_rec = &table->rarray[rid];
		IBPM_LOG(p_log, IBPM_ERR, "ERR: trid=0x%08lx lid=%d port="
				"%d type=%d timed out\n",
				 trid, p_rec->lid, p_rec->port, p_rec->type);
		return 1;
	}

	IBPM_LOG(p_log, IBPM_DEBUG, "got MAD with trid=0x%08lx at index=%d\n",
			trid, *free_id);

	/* decode the MAD */
	decode_mad(ibpm, mad, free_id, trid);

	return 0;
}

/*
* DESCRIPTION
*	This function closes the UDP
*	socket / HTTP connection.
*/
static void close_socket(ibpm_t *ibpm)
{
	if (!strcmp(ibpm->ufm_protocol, UFM_PROTOCOL_HTTP) || !strcmp(ibpm->ufm_protocol, UFM_PROTOCOL_HTTPS)) {
		if (ibpm->httphandle) {
			HTTPClient_cleanup(ibpm->httphandle);
			ibpm->httphandle = NULL;
		}
	}
	else {
		if (ibpm->sockfd) {
			close(ibpm->sockfd);
			ibpm->sockfd = 0;
		}
	}
}

/*
* DESCRIPTION
*	This function initialises the HTTP connection.
*/
static void init_http_connection(ibpm_t *ibpm, int use_ssl)
{
	sprintf(ibpm->ufm_rest_url, "%s://%s:%d%s", ibpm->ufm_protocol, ibpm->address, ibpm->port, ibpm->ufm_url);
	IBPM_LOG(&ibpm->log, IBPM_INFO, "Connecting %s on server %s via port %d; Full URL is: %s\n",
		ibpm->ufm_protocol, ibpm->address, ibpm->port, ibpm->ufm_rest_url);
	ibpm->httphandle = HTTPClient_Init(use_ssl);
}

/*
* DESCRIPTION
*	This function initializes the UDP
*	socket / HTTP connection.
*/
static void init_socket(ibpm_t *ibpm)
{
	if (!ibpm->update_socket) {
		return;
	}

	close_socket(ibpm);

	if (!strcmp(ibpm->ufm_protocol, UFM_PROTOCOL_HTTP)) {
		init_http_connection(ibpm, 0);
	}
	else if (!strcmp(ibpm->ufm_protocol, UFM_PROTOCOL_HTTPS)) {
		init_http_connection(ibpm, 1);
	}
	else {
		IBPM_LOG(&ibpm->log, IBPM_INFO, "Connecting %s on server %s via port %d\n", ibpm->ufm_protocol,
			ibpm->address, ibpm->port);
		ibpm->sockfd = socket(AF_INET, SOCK_DGRAM, 0);
		bzero(&ibpm->servaddr, sizeof(ibpm->servaddr));
		ibpm->servaddr.sin_family = AF_INET;
		ibpm->servaddr.sin_addr.s_addr = inet_addr(ibpm->address);
		ibpm->servaddr.sin_port = htons(ibpm->port);
	}

	ibpm->update_socket = 0;
}

/*
* DESCRIPTION
*  Initialization of the udp socket.
*/
static void init_header(ibpm_t *ibpm)
{
	header_t *header;

	header = &ibpm->message->header;
	strcpy(header->msg_class, "UFM\0");
	header->type = 0x5;
	header->version = 0x1;
	header->source_guid = 0xaa;
	header->source_port = 0x2;
}

/*
* DESCRIPTION
*	This function dump all
*	the statistics into a file.
*/
static void dump_statistics(ibpm_t *ibpm)
{
	FILE *stat_file;
	record_t *p_rec;
	struct timeval tv;
	tables_t *table = &ibpm->table[p_ibpm->active_tbl];
	stats_t *st = &ibpm->stats;
	log_t *p_log = &ibpm->log;
	int time_to_send = 0;
	time_t t = time(NULL);
	int i;

	ibpm->print_stats = 0;
	gettimeofday(&tv, NULL);

	stat_file = fopen(DEFAULT_STATS_FILENAME, "w+");
	if (!stat_file) {
		IBPM_LOG(p_log, IBPM_ERR,
				 "ERR: Cannot open stat file \'%s\':%m\n",
				 DEFAULT_STATS_FILENAME);
		return;
	}

	fprintf(stat_file, "ca = %s\nport = %d\nwindow size = %d\n"
			"rsize = %d\nssize = %d\ntable switches = %u\n"
			"message fragments = %u\nreset error counters = %u\n"
			"pc timeouts = %u\nvendor timeouts = %u\n"
			"socket timeouts = %u\ndropped = %u\n"
			"MAD send time:\n\tlast interval = %u(usec)\n"
			"\tmin avg = %u(usec)\n\tmax avg = %u(usec)\n",
			ibpm->ca, ibpm->ca_port, ibpm->window_size,
			table->rsize, table->ssize, st->switch_tbl,
			st->frag_msg, st->reset_error_pc,
			st->pc_timeouts, st->vendor_timeouts,
			st->socket_timeouts, st->dropped, st->avg_send_time,
			st->min_avg, st->max_avg);

	if (!ibpm->report_level)
		goto Exit;

	fprintf(stat_file, "%s", ctime(&t));
	fprintf(stat_file, "lid, port, devid sw_id time_to_send_msec\n");
	for (i = 0; i < table->rsize; i++) {
		p_rec = table->rlinks[i];
		time_to_send = diff_usec(&tv, &p_rec->send_time)/1000;
		fprintf(stat_file, "%d, %d, 0x%x, %d, %d\n",
				p_rec->lid, p_rec->port, p_rec->devid,
				p_rec->sw_id, time_to_send);
	}

Exit:
	fclose(stat_file);
}

static void
print_trace (FILE *stream)
{
	void  *array[50];
	size_t size;
	char  **strings;
	size_t  i;

	size = backtrace (array, 50);
	if (stream) {
		backtrace_symbols_fd (array, size, fileno(stream));
		fflush(stream);
		return;
	}

	strings = backtrace_symbols (array, size);

	fprintf (stderr, "Obtained %d stack frames.\n", (int)size);

	for (i = 0; i < size; i++) {
		fprintf(stderr, "%s\n", strings[i]);
		syslog(LOG_ALERT, "%s\n", strings[i]);
	}

	free (strings);
}

static void freeVlrArray(ibpm_t *ibpm, int psl)
{
	log_t *p_log = &ibpm->log;

	if (psl) {
		IBPM_LOG(p_log, IBPM_DEBUG,"Free lid_to_sl info.\n");
		free(ibpm->vlr_config.lid_to_sl);
		ibpm->vlr_config.lid_to_sl = NULL;
		free(ibpm->vlr_config.psl_file);
		ibpm->vlr_config.psl_file = NULL;
		ibpm->vlr_config.psl_mtime = 0;
	} else {
		IBPM_LOG(p_log, IBPM_DEBUG,"Free ca_to_sw_lid info.\n");
		free(ibpm->vlr_config.ca_to_sw_lid);
		ibpm->vlr_config.ca_to_sw_lid = NULL;
		free(ibpm->vlr_config.peer_path_file);
		ibpm->vlr_config.peer_path_file = NULL;
		ibpm->vlr_config.peer_path_mtime = 0;
	}
}

static int allocateVlrArray(void **buff, int size)
{
	*buff = malloc(size);
	if (! *buff) {
		return -1;
	} else {
		memset(*buff,0xFF,size);
	}
	return 0;
}

/*
* DESCRIPTION
*	free all allocated memory
*/
static void ibpm_exit(ibpm_t *ibpm)
{
	log_t *p_log = &ibpm->log;

	print_trace(p_log->file);

	/* dump statistics file before exiting */
	IBPM_LOG(p_log, IBPM_INFO, "dumping statistics\n");
	dump_statistics(ibpm);

	IBPM_LOG(p_log, IBPM_DEBUG, "free message\n");
	free(ibpm->message);
	IBPM_LOG(p_log, IBPM_DEBUG, "free trid array\n");
	free(ibpm->trid_arr);

	freeVlrArray(ibpm, 0);
	freeVlrArray(ibpm, 1);

	IBPM_LOG(p_log, IBPM_DEBUG, "free tables\n");
	free(ibpm->table[0].sarray);
	free(ibpm->table[0].rarray);
	free(ibpm->table[0].rlinks);
	free(ibpm->table[1].sarray);
	free(ibpm->table[1].rarray);
	free(ibpm->table[1].rlinks);

	IBPM_LOG(p_log, IBPM_DEBUG, "mad_rpc close port\n");
	mad_rpc_close_port(ibpm->srcport);

	/* close the UDP socket / HTTP connection if exists */
	close_socket(ibpm);

	/* close log file if opened */
	IBPM_LOG(p_log, IBPM_DEBUG, "close log file\n");
	if (p_log->file) {
		if (fclose(p_log->file))
			printf("cannot close logfile\n, %s %m\n", p_log->filename);
		p_log->file = NULL;
	}
}

static void parse_int(log_t *log, char *q, char *token, int *val, int dval)
{
	int  rc;

	q += strlen(token);
	rc = sscanf(q, "%d", val);
	if (rc != 1) {
		IBPM_LOG(log, IBPM_ERR, "ERR: invalid %s value - using default %d\n",
				token, dval);
		*val = dval;
	} else
		IBPM_LOG(log, IBPM_INFO, "%s = %d\n", token, *val);
}

static void parse_uint64(log_t *log, char *q, char *token, uint64_t *val,
		uint64_t dval)
{
	int  rc;

	q += strlen(token);
	rc = sscanf(q, "%llx", (long long unsigned int *) val);
	if (rc != 1) {
		IBPM_LOG(log, IBPM_ERR, "ERR: invalid %s value - using default %llx\n",
				token, dval);
		*val = dval;
	} else
		IBPM_LOG(log, IBPM_INFO, "%s = %llx\n", token, *val);
}

static void parse_string(log_t *log, char *q, char *token, char *val,
		char *dval)
{
	int  rc;

	q += strlen(token);
	rc = sscanf(q, "%s", val);
	if (rc != 1) {
		IBPM_LOG(log, IBPM_ERR, "ERR: invalid %s value - using default %s\n",
				token, dval);
		strcpy(val, dval);
	} else
		IBPM_LOG(log, IBPM_INFO, "%s = %s\n", token, val);
}

static void parse_file_path(log_t *log, char *q, char *token, char **val,
		char *dval)
{
	int  rc;
	char buff[1024];
	buff[0]='\0';

	q += strlen(token);
	rc = sscanf(q, "%s", buff);
	if (rc != 1) {
		IBPM_LOG(log, IBPM_ERR, "ERR: invalid %s value - using default %s\n",
				token, dval);
		if(*dval != '\0')
			strcpy(buff,dval);
	}

	if(*val) {
		if (strcmp(*val,buff) == 0) {
			IBPM_LOG(log, IBPM_INFO, "%s = %s (no change)\n", token, *val);
		} else {
			free(*val);
			*val = NULL;
		}
	}

	if((*buff != '\0') && (*val == NULL)){
		*val = strdup(buff);
		IBPM_LOG(log, IBPM_INFO, "%s = %s (value changed)\n", token, *val);
	}
	else if (*val == NULL)
		IBPM_LOG(log, IBPM_INFO, "%s = NULL\n", token);

}

static void setRereadVlrFiles(ibpm_t *ibpm)
{
	ibpm->vlr_config.psl_mtime = 0;
	ibpm->vlr_config.peer_path_mtime = 0;
}

// Parse Peer Path file and set the SLT tables accordingly
/*
Channel Adapter 0x0000000000100001, base LID 3, LMC 0, "H_0_0_0_0", port 1
# LID  : MTU : RATE
0x0002 : 3   : 3
*/
static int parsePeerPathFile(ibpm_t *ibpm)
{
	char line[1024];
	int rc;
	log_t *p_log = &ibpm->log;
	char *p;
	FILE *file;
	uint64_t source_guid;
	unsigned int ca_lid;
	unsigned int sw_lid;

	char * ca_token = "Channel Adapter ";
	int ca_token_size = strlen(ca_token);
	char * header_token = "# LID  :";
	int header_token_size = strlen(header_token);

	file = fopen(ibpm->vlr_config.peer_path_file, "r");
	if (!file) {
		IBPM_LOG(p_log, IBPM_ERR,
				 "ERR: Cannot open peer path file \'%s\': %m\n",
				 ibpm->vlr_config.peer_path_file);
		return -1;
	} else
		IBPM_LOG(p_log, IBPM_INFO, "file %s successfully opened\n",
				 ibpm->vlr_config.peer_path_file);

	if (ibpm->vlr_config.ca_to_sw_lid == NULL) {
		if (allocateVlrArray((void**)(&ibpm->vlr_config.ca_to_sw_lid),
							 VLR_ARRAY_SIZE*sizeof(uint16_t))) {
			IBPM_LOG(p_log, IBPM_ERR,
					 "ERR: Could not allocate memory ca_to_sw_lid \n");
			goto Err;
		}
	} else
		memset(ibpm->vlr_config.ca_to_sw_lid, 0xFF, VLR_ARRAY_SIZE*sizeof(uint16_t));

	while (fgets(line, sizeof(line) - 1, file) != NULL) {

		//CA line
		if (strncmp(ca_token, line, ca_token_size))
			continue;

		p = line+ca_token_size;
		rc = sscanf(p, "%llx, base LID %u",
					(long long unsigned int *)(&source_guid), &ca_lid);
		if (rc < 2 ||ca_lid < 1 || ca_lid > MAX_UCAST_LID) {
			IBPM_LOG(p_log, IBPM_ERR,
					 "ERR: invalid CA record format read failed rc=%d in \'%s\'\n",
					 rc,line);
			continue;
		}

		//Header
		if (fgets(line, sizeof(line) - 1, file) == NULL) {
			IBPM_LOG(p_log, IBPM_ERR, "ERR: unexpected EOF instead of header.\n");
			break;
		}
		if (strncmp(header_token, line, header_token_size)) {
			IBPM_LOG(p_log, IBPM_ERR, "ERR: expecting \'%s\' instead of \'%s\'.\n",
					 header_token, line);
			continue;
		}

		//SW lid
		if (fgets(line, sizeof(line) - 1, file) == NULL) {
			IBPM_LOG(p_log, IBPM_ERR, "ERR: unexpected EOF instead of SW info.\n");
			break;
		}
		rc = sscanf(line, "%x", &sw_lid);
		if (rc < 1 ||sw_lid < 1 || sw_lid > MAX_UCAST_LID) {
			IBPM_LOG(p_log, IBPM_ERR,
					 "ERR: invalid SW record format read failed rc=%d in \'%s\'\n",
					 rc,line);
			continue;
		}

		IBPM_LOG(p_log, IBPM_DEBUG,
				 "parsePeerPathFile: CA lid:%d(0x%x) SW lid:%d(0x%x).\n",
				 ca_lid,ca_lid,sw_lid,sw_lid);
		ibpm->vlr_config.ca_to_sw_lid[ca_lid] = (uint16_t)sw_lid;

	}

	fclose(file);
	return 0;

Err:
	fclose(file);
	return -1;
}


// Parse PSL file and set the SLT tables accordingly
/*
version: 1.0.0
routing_engine: torus-2QoS

Switch 0x0000000000200000, base LID 2, "S_0_0_0", port 0
# LID  : SL : MTU : RATE
0x0002 : 0  : 3   : 3
0x0004 : 0  : 3   : 3
0x0018 : 0  : 3   : 3

*/
static int parsePSLFile(ibpm_t *ibpm)
{
	char line[1024];
	int rc;
	log_t *p_log = &ibpm->log;
	char *p;
	FILE *file;
	uint64_t source_guid;
	unsigned int lid;
	unsigned int sw_lid;
	unsigned int sl;
	uint16_t baselid = ibpm->baselid;

	char * ver_token = "version: ";
	int ver_token_size = strlen(ver_token);
	char * re_token = "routing_engine: ";
	int re_token_size = strlen(re_token);
	char * sw_token = "Switch ";
	int sw_token_size = strlen(sw_token);
	char * ca_token = "Channel Adapter ";
	int ca_token_size = strlen(ca_token);
	char * header_token = "# LID  :";
	int header_token_size = strlen(header_token);
	int is_torus = 0;
	int read_sl_info = 0;
	int iter;

	if (ibpm->baselid == 0) {
		IBPM_LOG(p_log, IBPM_ERR,"ERR: Invalid baselid 0\n");
		return -1;
	}

	file = fopen(ibpm->vlr_config.psl_file, "r");
	if (!file) {
		IBPM_LOG(p_log, IBPM_ERR,
				 "ERR: Cannot open psl path file \'%s\': %m\n",
				 ibpm->vlr_config.psl_file);
		return -1;
	} else
		IBPM_LOG(p_log, IBPM_INFO, "file %s successfully opened\n",
				 ibpm->vlr_config.psl_file);

	if (ibpm->vlr_config.lid_to_sl == NULL) {
		if (allocateVlrArray((void**)(&ibpm->vlr_config.lid_to_sl),
							 VLR_ARRAY_SIZE*sizeof(uint8_t))) {
			IBPM_LOG(p_log, IBPM_ERR,
					 "ERR: Could not allocate memory lid_to_sl \n");
			goto Err;
		}
	} else
		memset(ibpm->vlr_config.lid_to_sl, 0xFF, VLR_ARRAY_SIZE*sizeof(uint8_t));

	/*
	version: 1.0.0
	routing_engine: torus-2QoS
	*/
	while (fgets(line, sizeof(line) - 1, file) != NULL) {

		if (*line == '\n')
			continue;

		if (strncmp(ver_token, line, ver_token_size)) {
			/*Full path sl file*/
			rewind(file);
			break;
		}
		p = line+ver_token_size;
		/*last ver number is for modification that do not break parsing*/
		if (strncmp("1.0.",p,strlen("1.0."))) {
			IBPM_LOG(p_log, IBPM_ERR,
					 "ERR: Unsupported psl file version:%s \n", p);
			goto Err;
		}
		if (fgets(line, sizeof(line) - 1, file) == NULL) {
			IBPM_LOG(p_log, IBPM_ERR, "ERR: unexpected EOF instead of PSL info.\n");
			goto Err;
		}

		if (strncmp(re_token, line, re_token_size)) {
			rewind(file);
			break;
		}

		p = line+re_token_size;
		if (strncmp("torus-2QoS",p,strlen("torus-2QoS"))) {
			IBPM_LOG(p_log, IBPM_ERR,
					 "ERR: Unsupported psl file routing_engine:%s \n", p);
			goto Err;
		}
		if (ibpm->vlr_config.ca_to_sw_lid == NULL) {
			IBPM_LOG(p_log, IBPM_ERR,
					 "No CA to SW info available. "
					 "Peer path file should be supplied if routing_engine is torus-2QoS \n");
			goto Err;
		}
		if (ibpm->vlr_config.ca_to_sw_lid[ibpm->baselid] == 0xFFFF) {
			IBPM_LOG(p_log, IBPM_ERR,
					 "ERR: No data for peer sw lid for base lid:%d \n", ibpm->baselid);
			goto Err;
		}
		baselid = ibpm->vlr_config.ca_to_sw_lid[ibpm->baselid];
		IBPM_LOG(p_log, IBPM_DEBUG,
				 "Parse PSL file in torus-2QoS format. sw lid:%d\n",baselid);
		is_torus = 1;
		break;
	}

	//look for baselid
	rc = 0;
	lid = 0;
	while (fgets(line, sizeof(line) - 1, file) != NULL) {

		if (*line == '\n' || *line == '0' || *line == '#')
			continue;


		if (!is_torus &&
			(strncmp(ca_token, line, ca_token_size) == 0)) {

			p = line+ca_token_size;
			rc = sscanf(p, "%llx, base LID %u",
						(long long unsigned int *)(&source_guid), &lid);

		} else if (strncmp(sw_token, line, sw_token_size) == 0) {

			p = line+sw_token_size;
			rc = sscanf(p, "%llx, base LID %u",
						(long long unsigned int *)(&source_guid), &lid);
		} else {
			continue;
		}

		if (rc != 2) {
			IBPM_LOG(p_log, IBPM_ERR,
					 "ERR: invalid record format read failed rc=%d in \'%s\'\n",
					 rc,line);
			continue;
		}
		if (lid == baselid) {
			read_sl_info = 1;
			break;
		}
	}

	if (!read_sl_info) {
		IBPM_LOG(p_log, IBPM_ERR, "ERR: Failed to find lid:%d info in psl file.\n",
				 baselid);
		goto Err;
	}

	//read header info
	if (fgets(line, sizeof(line) - 1, file) == NULL) {
		IBPM_LOG(p_log, IBPM_ERR, "ERR: unexpected EOF instead of \'%s ...\'.\n",
				 header_token);
		goto Err;
	}
	if (strncmp(header_token, line, header_token_size)) {
		IBPM_LOG(p_log, IBPM_ERR, "ERR: expecting \'%s\' instead of \'%s\'.\n",
				 header_token, line);
		goto Err;
	}

	//read SL info
	while (fgets(line, sizeof(line) - 1, file) != NULL) {
		if (*line != '0')
			break;

		rc = sscanf(line, "%x : %d", &lid, &sl);
		if (rc < 2 ||lid < 1 || lid > MAX_UCAST_LID || sl > 15) {
			IBPM_LOG(p_log, IBPM_ERR,
					 "ERR: invalid SW record format read failed rc=%d in \'%s\'\n",
					 rc,line);
			continue;
		}
		IBPM_LOG(p_log, IBPM_DEBUG,
				 "parsePSLFile: lid:%d(0x%x) sl:%u.\n", lid,lid,sl);
		ibpm->vlr_config.lid_to_sl[lid] = (uint8_t)sl;
	}

	if (is_torus) {
		/*we know all SW sl's set them to all connected CA's*/
		for (iter = 1; iter < VLR_ARRAY_SIZE; ++iter) {
			sw_lid = ibpm->vlr_config.ca_to_sw_lid[iter];
			if (sw_lid == 0xFFFF)
				continue; /* not ca lid*/
			if (ibpm->vlr_config.lid_to_sl[sw_lid] == 0xFF) {
				IBPM_LOG(p_log, IBPM_ERR,
						 "ERR: No SL info for sw lid:%u\n", sw_lid);
				continue;
			}
			ibpm->vlr_config.lid_to_sl[iter]  =
			ibpm->vlr_config.lid_to_sl[sw_lid];
		}
	}

	fclose(file);
	return 0;

Err:
	fclose(file);
	return -1;
}

/*
* DESCRIPTION
*	This function reads the configuration
*	file and set variables accordingly.
*/
static void read_configuration(ibpm_t *ibpm)
{
	char line[1024];
	FILE *conf_file;
	log_t *p_log = &ibpm->log;
	char *q, *p;
	int checkdebug = 1;
	int checkverbose = 1;
	int checkreport = 1;
	int checkcsv = 1;
	int checkportretries = 1;
	int checkportinterval = 1;

	ibpm->read_conf = 0;
	conf_file = fopen(DEFAULT_CONF_FILENAME, "r");
	if (!conf_file) {
		IBPM_LOG(p_log, IBPM_WARN,
				 "ERROR; Cannot open stat file \'%s\':%m\n",
				 DEFAULT_CONF_FILENAME);
		return;
	}

	IBPM_LOG(p_log, IBPM_ERR, "read_configuration file: %s \n",DEFAULT_CONF_FILENAME);

	while (fgets(line, sizeof(line) - 1, conf_file) != NULL) {
		p = line;

		/* skip rest of the line after '#' */
		q = strchr(p, '#');
		if (q)
			*q = '\0';

		if (checkdebug && (q = strstr(p, "debug"))) {
			parse_int(p_log, q, "debug", &ibpm->log.dbg_lvl, 1);
			checkdebug = 0;
			continue;
		}

		if (checkverbose && (q = strstr(p, "verbose"))) {
			parse_int(p_log, q, "verbose", &ibpm->verbose, 0);
			checkverbose = 0;
			continue;
		}

		if (checkcsv && (q = strstr(p, "csv"))) {
			parse_int(p_log, q, "csv", &ibpm->csv, 0);
			checkcsv = 0;
			continue;
		}

		if (checkreport && (q = strstr(p, "report"))) {
			parse_int(p_log, q, "report", &ibpm->report_level, 0);
			checkreport = 0;
			continue;
		}

		if (checkportretries && (q = strstr(p, "port_poll_retries"))) {
			parse_int(p_log, q, "port_poll_retries", &ibpm->port_poll_retries,
					DEFAULT_POLL_RETRIES);
			checkportretries = 0;
			continue;
		}

		if (checkportinterval && (q = strstr(p, "port_poll_interval"))) {
			parse_int(p_log, q, "port_poll_interval", &ibpm->port_poll_interval,
					DEFAULT_POLL_INTERVAL);
			checkportinterval = 0;
			continue;
		}
	}
	fclose(conf_file);
}

static void read_conf(int signo)
{
	IBPM_LOG(&p_ibpm->log, IBPM_ERR, "Reading conf file"
			 " (%s)\n", DEFAULT_CONF_FILENAME);
	p_ibpm->read_conf = 1;
}

/*
* DESCRIPTION
*	This function reset the
*	gathered statistics.
*/
static void reset_statistics(ibpm_t *ibpm)
{
	memset(&ibpm->stats, 0, sizeof(ibpm->stats));
	ibpm->reset_stats = 0;
}

static void reset_stats(int signo)
{
	IBPM_LOG(&p_ibpm->log, IBPM_ERR,
			 "Reset stats\n");
	p_ibpm->reset_stats = 1;
}

static void print_stats(int signo)
{
	IBPM_LOG(&p_ibpm->log, IBPM_ERR,
			 "Dumping stats file"
			 " (%s)\n", DEFAULT_STATS_FILENAME);
	p_ibpm->print_stats = 1;

}

/*
* DESCRIPTION
*	catch signal and exit.
*/
static void catch_signal(int signo)
{
	IBPM_LOG(&p_ibpm->log, IBPM_ERR,
			 "caught signal %s(%d); Exiting...\n",
			 strsignal(signo), signo);

	switch (signo) {
	case SIGSEGV:
	case SIGBUS:
	case SIGFPE:
	case SIGTERM:
	case SIGINT:
		ibpm_exit(p_ibpm);
		exit(1);
		break;
	default:
		break;
	}
}

/*
* DESCRIPTION
*	register signals to catch.
*/
static void register_signal(void)
{
	struct sigaction act;

	memset(&act,0,sizeof(sigaction));

	sigemptyset(&act.sa_mask);
	act.sa_handler = catch_signal;
	act.sa_flags = 0;
	sigaction(SIGSEGV, &act, NULL);
	sigaction(SIGBUS, &act, NULL);
	sigaction(SIGFPE, &act, NULL);
	sigaction(SIGTERM, &act, NULL);
	sigaction(SIGINT, &act, NULL);

	act.sa_handler = print_stats;
	sigaction(SIGUSR1, &act, NULL);

	act.sa_handler = reset_stats;
	sigaction(SIGUSR2, &act, NULL);

	act.sa_handler = read_conf;
	sigaction(SIGHUP, &act, NULL);
}

/*
* DESCRIPTION
*	open log and config files
*	if necessary.
*/
static void ibpm_open_files(ibpm_t *ibpm)
{
	log_t *p_log = &ibpm->log;
	char *log_dir = "/var/log/ibpm";
	char *csv_dir = "/var/log/ibpm/csv";

	/* create ibpm log directories */
	mkdir(log_dir, 755);
	mkdir(csv_dir, 755);

	/* open log file*/
	if (p_log->filename) {
		if (ibpm->eraselog)
			p_log->file = fopen(p_log->filename, "w+");
		else
			p_log->file = fopen(p_log->filename, "a+");
		if (!p_log->file)
			IBPM_LOG(p_log, IBPM_ERR,
					 "ERR: Cannot open log file \'%s\':%m\n",
					 p_log->filename);
		else {
			setlinebuf(p_log->file);
			IBPM_LOG(p_log, IBPM_INFO, "%s opened successfully\n",
					 p_log->filename);
		}
	}
}

/*
* DESCRIPTION
*	set default values.
*/
static void ibpm_default_init(ibpm_t *ibpm)
{
	ibpm->window_size = WINDOW_SIZE;
	ibpm->stats.min_avg = 99999;
	ibpm->report_level = 0;
	ibpm->read_conf = 0;
	ibpm->print_stats = 0;
	ibpm->reset_stats = 0;
	ibpm->ca = 0;
	ibpm->ca_port = 0;
	ibpm->csv = 0;
	ibpm->eraselog = 0;
	ibpm->timeout = DEFAULT_TIMEOUT;
	ibpm->leasetime = 0;
	ibpm->numrecords = MAX_SIZE;
	ibpm->finished = 0;
	ibpm->log.dbg_lvl = 0;
	ibpm->log.file = 0;
	ibpm->log.csvfile = 0;
	ibpm->config.file = 0;
	ibpm->config.mtime = 0;
	ibpm->log.filename = DEFAULT_LOG_FILENAME;
	ibpm->config.filename  = DEFAULT_REC_FILENAME;
	ibpm->active_tbl = 0;
	ibpm->verbose = 0;
	ibpm->errrate = 0;
	ibpm->update_socket = 1;
	ibpm->default_poll_time = DEFAULT_SAMPLE_INTERVAL;
	ibpm->default_poll_time_msec = ibpm->default_poll_time * 1000;
	ibpm->reset_mode = RESET_EVERY_POLL;
	ibpm->threshold = DEFAULT_THRESHOLD;
	ibpm->port_poll_interval = DEFAULT_POLL_INTERVAL;
	ibpm->port_poll_retries = DEFAULT_POLL_RETRIES;
	ibpm->max_timeouts_number = TIMEOUTS_MUX_NUMBER;
	ibpm->baselid = 0;

	ibpm->vlr_config.psl_file = NULL;
	ibpm->vlr_config.psl_cmdln = 0;
	ibpm->vlr_config.psl_mtime = 0;
	ibpm->vlr_config.psl_rc = 0;
	ibpm->vlr_config.lid_to_sl = NULL;

	ibpm->vlr_config.peer_path_file = NULL;
	ibpm->vlr_config.peer_cmdln = 0;
	ibpm->vlr_config.peer_path_mtime = 0;
	ibpm->vlr_config.peer_rc = 0;
	ibpm->vlr_config.ca_to_sw_lid = NULL;

	ibpm->sockfd = 0;
	ibpm->httphandle = NULL;
}


/*
* DESCRIPTION
*	init pc_threshold structure
*/
static void init_threshold(ibpm_t *ibpm)
{
	struct perf_count *pc_thr = &ibpm->pc_thr;
	log_t *p_log = &ibpm->log;
	unsigned int threshold = ibpm->threshold;

	pc_thr->symbolerrors = (threshold * 65535) / 100;
	pc_thr->linkrecovers = (threshold * 255) / 100;
	pc_thr->linkdowned = (threshold * 255) / 100;
	pc_thr->rcverrors = (threshold * 65535) / 100;
	pc_thr->rcvremotephyerrors = (threshold * 65535) / 100;
	pc_thr->rcvswrelayerrors = (threshold * 65535) / 100;
	pc_thr->xmtdiscards = (threshold * 65535) / 100;
	pc_thr->xmtconstrainterrors = (threshold * 255) / 100;
	pc_thr->rcvconstrainterrors = (threshold * 255) / 100;
	pc_thr->linkintegrityerrors = (threshold * 15) / 100;
	pc_thr->excbufoverrunerrors = (threshold * 15) / 100;
	pc_thr->vl15dropped = (threshold * 65535) / 100;
	pc_thr->xmtwait = (threshold * 4294967295u) / 100;

	/* if reset mode == RESET_EVERY_POLL then
	 * we need to set the data counter's threshold
	 * to be zero (this way we will reset them every poll).
	 */
	if (ibpm->reset_mode == RESET_EVERY_POLL)
		threshold = 0;

	pc_thr->rcvdata = (threshold * 4294967295u) / 100;
	pc_thr->rcvpkts = (threshold * 4294967295u) / 100;
	pc_thr->xmtdata = (threshold * 4294967295u) / 100;
	pc_thr->xmtpkts = (threshold * 4294967295u) / 100;

	if (ibpm->verbose)
		IBPM_LOG(p_log, IBPM_ERR, "counters thresholds:\n"
				"symbolerrors=%d\nlinkrecovers=%d\nlinkdowned=%d\n"
				"rcverrors=%d\nrcvremotephyerrors=%d\nrcvswrelayerrors=%d\n"
				"xmtdiscards=%d\nxmtconstrainterrors=%d\nrcvconstrainterrors=%d\n"
				"linkintegrityerrors=%d\nexcbufoverrunerrors=%d\nvl15dropped=%d\n"
				"xmtwait=%lld\nrcvdata=%lld\nrcvpkts=%lld\nxmtdata=%lld\n"
				"xmtpkts=%lld\n",
				pc_thr->symbolerrors, pc_thr->linkrecovers, pc_thr->linkdowned,
				pc_thr->rcverrors, pc_thr->rcvremotephyerrors,
				pc_thr->rcvswrelayerrors, pc_thr->xmtdiscards,
				pc_thr->xmtconstrainterrors, pc_thr->rcvconstrainterrors,
				pc_thr->linkintegrityerrors, pc_thr->excbufoverrunerrors,
				pc_thr->vl15dropped, pc_thr->xmtwait, pc_thr->rcvdata,
				pc_thr->rcvpkts, pc_thr->xmtdata, pc_thr->xmtpkts);
}

/* if fails Exit ibpm */
static void ibpm_increase_numrecords(ibpm_t *ibpm, int ports_to_add)
{
	log_t *p_log = &ibpm->log;
	tables_t tmp_table;
	int numrecords;

	if ( ports_to_add == 0 )
		numrecords = ibpm->table[!ibpm->active_tbl].numrecords;
	else {
		/* add 5% ports and multiply by 2 for ext counters*/
		numrecords = (int)((ibpm->table[ibpm->active_tbl].numrecords + ports_to_add*2)*1.05);

		if ( numrecords < ibpm->numrecords )
			numrecords = ibpm->numrecords;

		if ( numrecords < ibpm->table[!ibpm->active_tbl].numrecords )
			numrecords = ibpm->table[!ibpm->active_tbl].numrecords;
	}

	if ( ibpm->table[ibpm->active_tbl].numrecords >= numrecords ) {
		IBPM_LOG(p_log, IBPM_ERR,
				 "ERR: Will not decrease rarray num records from %d to %d. \n",
				 ibpm->table[ibpm->active_tbl].numrecords, numrecords);
		return;
	}

	IBPM_LOG(p_log, IBPM_DEBUG, "Increase numrecords from %d to %d (table:%d). \n",
			 ibpm->table[ibpm->active_tbl].numrecords, numrecords, ibpm->active_tbl);

	tmp_table.rarray = (record_t *) calloc(numrecords, sizeof(record_t));
	tmp_table.rlinks = calloc(numrecords, sizeof(record_t *));

	if ( tmp_table.rarray == NULL || tmp_table.rlinks == NULL ) {
		IBPM_LOG(p_log, IBPM_ERR, "ERR: Could not increase rarray num records to %d. \n",numrecords);
		syslog(LOG_ALERT,"ERR: Could not increase rarray num records to %d. \n",numrecords);
		free(tmp_table.rarray);
		free(tmp_table.rlinks);
		raise(SIGTERM); /* Exit ibpm if failed to allocate memory*/
	} else {
		free(ibpm->table[ibpm->active_tbl].rarray);
		free(ibpm->table[ibpm->active_tbl].rlinks);
		ibpm->table[ibpm->active_tbl].rarray = tmp_table.rarray;
		ibpm->table[ibpm->active_tbl].rlinks = tmp_table.rlinks;
		ibpm->table[ibpm->active_tbl].numrecords = numrecords;
	}

	/* On startup allocate the inactive tbl */
	if (ibpm->table[!ibpm->active_tbl].rarray == NULL) {

		IBPM_LOG(p_log, IBPM_DEBUG, "Increase numrecords from %d to %d (table:%d). \n",
				 ibpm->table[!ibpm->active_tbl].numrecords, numrecords, !ibpm->active_tbl);

		ibpm->table[!ibpm->active_tbl].rarray = (record_t *) calloc(numrecords, sizeof(record_t));
		ibpm->table[!ibpm->active_tbl].rlinks = calloc(numrecords, sizeof(record_t *));
		ibpm->table[!ibpm->active_tbl].numrecords = numrecords;

		if ( ibpm->table[!ibpm->active_tbl].rarray == NULL ||
			 ibpm->table[!ibpm->active_tbl].rlinks == NULL ) {
			IBPM_LOG(p_log, IBPM_ERR, "ERR: Could not cllocate %d records. \n",numrecords);
			syslog(LOG_ALERT,"ERR: Could not cllocate %d records. \n",numrecords);
			raise(SIGTERM); /* Exit ibpm if failed to allocate memory*/
		}
	}
	return;
}

/*
* DESCRIPTION
*	init ibpm structure
*/
static void ibpm_init(ibpm_t *ibpm)
{
	log_t *p_log = &ibpm->log;
	config_t *p_config = &ibpm->config;
	time_t t;

	int mgmt_classes[] = {IB_SMI_DIRECT_CLASS, IB_PERFORMANCE_CLASS, IB_MLX_VENDOR_CLASS};
        ibpm->numrecords *= 2; /* for ext records */

#define NUM_CLASSES ARR_SIZE(mgmt_classes)

	/* register signal */
	register_signal();

	/* mad_rpc open port */
	ibpm->srcport = mad_rpc_open_port(ibpm->ca, ibpm->ca_port, mgmt_classes, NUM_CLASSES);
	if (!ibpm->srcport) {
		syslog(LOG_ALERT, "ERR: Failed to open '%s' port '%d'\n",
				ibpm->ca, ibpm->ca_port);
		exit(1);
	}

	ibpm->message = (message_t *) calloc(1, sizeof(message_t));
	ibpm->trid_arr = (trid_arr_t *) calloc(ibpm->window_size, sizeof(trid_arr_t));

	ibpm->table[0].sarray = (sw_rec_t *) calloc(MAX_SW_SIZE, sizeof(sw_rec_t));
	ibpm->table[0].rarray = NULL;
	ibpm->table[0].rlinks = NULL;
	ibpm->table[1].rsize = 0;
	ibpm->table[0].numrecords = 0;
	ibpm->table[1].sarray = (sw_rec_t *) calloc(MAX_SW_SIZE, sizeof(sw_rec_t));
	ibpm->table[1].rarray = NULL;
	ibpm->table[1].rlinks = NULL;
	ibpm->table[1].rsize = 0;
	ibpm->table[1].numrecords = 0;

	/* allocate .rarray, .rlinks and set numrecords on parse_record
           after counting ports number.
	ibpm->table[0].rarray = (record_t *) calloc(numrecords, sizeof(record_t));
	ibpm->table[0].rlinks = calloc(numrecords, sizeof(record_t *));
	ibpm->table[1].rarray = (record_t *) calloc(numrecords, sizeof(record_t));
	ibpm->table[1].rlinks = calloc(numrecords, sizeof(record_t *));
	*/

	if (ibpm->table[0].sarray == NULL ||
	    ibpm->table[1].sarray == NULL ) {

		IBPM_LOG(p_log, IBPM_ERR, "ERR: Could not allocate memory, Exiting...\n");
		syslog(LOG_ALERT,"ERR: Could not allocate memory, Exiting...\n");
		exit(1);
	}

	/* make header */
	init_header(ibpm);

	/* init log files */
	ibpm_open_files(ibpm);

	t = time(NULL);
	IBPM_LOG(p_log, IBPM_ERR, "Started on %s\n", ctime(&t));
	/* print options */
	IBPM_LOG(p_log, IBPM_ERR, "config file \'%s\'.\n",
			 p_config->filename);
	IBPM_LOG(p_log, IBPM_ERR, "minimal initial number of record %d\n",
			 ibpm->numrecords);
	IBPM_LOG(p_log, IBPM_ERR, "dbg_lvl = %d\n", p_log->dbg_lvl);
	IBPM_LOG(p_log, IBPM_ERR, "version = %s\n", IBPM_VERSION);
	IBPM_LOG(p_log, IBPM_ERR, "window = %d\n", ibpm->window_size);
	IBPM_LOG(p_log, IBPM_ERR, "ca = %s\n", ibpm->ca);
	IBPM_LOG(p_log, IBPM_ERR, "ca_port = %d\n", ibpm->ca_port);

	/* deamonize ibpm if not running on foreground */
	if (p_log->filename) {
		if (daemon(0, 0)) {
			IBPM_LOG(p_log, IBPM_ERR, "ERR: Could not daemonize process, Exiting...");
			exit(1);
		}
	}

	/* init and read extended_pc_not_supported_devids */
	memset(ibpm->epc_support_info,0,sizeof(ibpm->epc_support_info));
	ibpm->epc_support_info_size = 0;
}

/*
* DESCRIPTION
* print usage to screen
*/
static void show_usage(void)
{
	printf("\nUsage: ibpm [-c(onf) <records_file> -t(imeout) <time_ms>"
		   " -l(og) <log_file> -f(oreground) -(x)csv -n(um) <max_num_rec>"
		   " -V(ersion) -d(ebug) <level> -C <ca_name> -P <ca_port> -z -h(elp)"
		   " -s(path_sl_file) <file> -p(peer_path_file) <file> ]\n");
	printf("Options:\n"
		   "--conf, -c <records_file>\tSpecify an alternative records file.\n"
		   "          \t\t\tDefault is '%s'.\n"
		   "--timeout, -t <time_ms>\t\tSpecify timeout (ms) for each MAD.\n"
		   "          \t\t\tDefault is %d ms.\n"
		   "--log, -l <log_file>\t\tSpecify an alternative log file.\n"
		   "          \t\t\tDefault is '%s'.\n"
		   "--foreground, -f\t\tWrite log output to the foregournd.\n"
		   "--csv, -x\t\t\tExport output statistics to\n"
		   "          \t\t\tcsv files under '%s'.\n"
		   "--num, -n <max_num_rec>\t\tSet the maximum number of monitored ports.\n"
		   "          \t\t\tDefault is %d.\n"
		   "--version, -V\t\t\tShow Version.\n"
		   "--debug, -d <level>\t\tThis option specifies a debug option.\n"
		   "          \t\t\tThe number following -d selects the debug\n"
		   "          \t\t\toption to enable\n"
		   "          \t\t\tLevel   Description\n"
		   "          \t\t\t-----  -----------------\n"
		   "          \t\t\t0  - Only Error messages - Default debug level\n"
		   "          \t\t\t1  - Warning messages\n"
		   "          \t\t\t2  - Info messages\n"
		   "          \t\t\t3  - Debug messages\n"
		   "--verbose, -v\t\t\tPrint general info messages.\n"
		   "--eraselog, -e\t\t\tDelete old log file. By default log file is accumulative\n"
		   "--C, -C <ca_name>\t\tSpecify ca_name.\n"
		   "--P, -P <ca_port>\t\tSpecify ca_port.\n"
		   "     -z\t\t\t\tPrint template conf file.\n"
		   "--pollretries, -R <retries>\tSpecify number of polling retries if port is down before shutdown\n"
		   "--max_timeouts, -m <number>\tSpecify number of polling retries when there is a timeout while sending MAD\n"
		   "--polltimeout, -T <timeout>\tSpecify polling timeout per retry when port is down before shutdown\n"
		   "--path_sl_file, -s <file>\tSpecify path-records.dump file path generated by dump_pr SM plugin\n"
		   "--peer_path_file, -p <file>\tSpecify opensm-peer-paths.dump file path generated by dump_pr SM plugin\n"
		   "--help, -h, -?\t\t\tDisplay this usage info then exit.\n\n"
		   "Configuration Files:\n"
		   "     \t\t\t"DEFAULT_CONF_FILENAME":\n"
		   "          \t\t\tIBPM conf file. Use -z option to print template.\n"
		   "          \t\t\tThe file is loaded/reloaded only after receiving SIGHUP signal.\n\n",
		   DEFAULT_REC_FILENAME, DEFAULT_TIMEOUT, DEFAULT_LOG_FILENAME,
		   DEFAULT_CSV_FILENAME, MAX_SIZE);
	fflush(stdout);
}

/*
* DESCRIPTION
*	return 0 if a device id has a unique entry
*	in the ib_to_hw matrix.
*	If not, we need to send gi (general info)
*	MAD to find the device's revision.
*/
static int check_dev_rev(tables_t *table, int id)
{
	int i;
	int count = 0;
	int devid = table->sarray[id].devid;

	for (i = 0; i < NUM_HW_DEV; i++) {
		if (devid == (ib_to_hw[i][0] & 0xffff)) {
			table->sarray[id].rev = ib_to_hw[i][0] >> 16;
			count++;
			if (count == 2)
				return 1;
		}
	}
	return 0;
}


/*
* DESCRIPTION
*	send gi (general info) MAD to
*	obtain the device's revision.
*/
static int get_sw_rev(ibpm_t *ibpm, tables_t *table, int id)
{
	ib_portid_t portid = {0};
	char buf[1024];
	is3_general_info_t *gi;
	ib_vendor_call_t call;

	memset(&buf, 0, sizeof(buf));

	gi = (is3_general_info_t *)&buf;

	portid.lid = table->sarray[id].lid;
	if (ibpm->vlr_config.lid_to_sl && setMadSl(ibpm, &portid))
		return -1;

	call.attrid = IB_MLX_IS3_GENERAL_INFO;
	call.mgmt_class = IB_MLX_VENDOR_CLASS;
	call.method = IB_MAD_METHOD_GET;
	call.timeout = 50; /* msec */

	if (!ib_vendor_call_via(&buf, &portid, &call, ibpm->srcport))
		return -1;

	table->sarray[id].rev = ntohs(gi->hw_info.hw_revision) >> 8;
	IBPM_LOG(&ibpm->log, IBPM_INFO, "lid=%d devid=0x%x rev=%d\n",
			 table->sarray[id].lid, table->sarray[id].devid, table->sarray[id].rev);
	return 0;
}

/*
* DESCRIPTION
*	run over all the switch table
*	and check for the revision.
*/
static int check_sw_rev(ibpm_t *ibpm, tables_t *table)
{
	int i;

	for (i = 0; i < table->ssize; i++) {
		if (table->sarray[i].rev != 0) {
			IBPM_LOG(&ibpm->log, IBPM_DEBUG, "sw id=%d already has rev(0x%x)\n", i, table->sarray[i].rev);
			continue;
		}
		if (check_dev_rev(table, i))
			get_sw_rev(ibpm, table, i);
	}
	return 0;
}

/*
* DESCRIPTION
*	init the time for all the records
*	which run for the first time -
*	(rec->firstrun = 1)
*/
static void init_rec_time(tables_t *table, struct timeval *curtime)
{
	int i = 0;
	record_t *p_rec = 0;

	for (i = 0; i < table->rsize; i++) {
		p_rec = &table->rarray[i];
		if (!p_rec->firstrun)
			continue;
		p_rec->isample = p_rec->ireport;
		p_rec->last_time = *curtime;
		p_rec->send_time = *curtime;
	}
}

static int send_record(ibpm_t *ibpm, record_t *p_rec,
		struct timeval *curtime, int index)
{
	/* update_rec_time */
	update_rec_time(p_rec, curtime);
	if (p_rec->type == MAD_TYPE_PC) {
		if (pc_mad_get(ibpm, p_rec, index, FALSE) < 0)
			return -1;
	} else if (p_rec->type == MAD_TYPE_PC_EXT) {
		if (pc_mad_get(ibpm, p_rec, index, TRUE) < 0)
			return -1;
	} else if (encode_xmitwait(ibpm, p_rec, index) < 0)
		return -1;
	return 0;
}

/* get the number of rec MAD before refresh local time */
/* alow accuracy of 0.1 sec */
int calc_refresh_time_count(ibpm_t *ibpm)
{
	int timeouts_per_sec = 0;
	int refresh_time_count = 0;

	if (ibpm->timeout == 0)
		timeouts_per_sec = 1;
	else {
		/* 100 is for accuracy of 0.1 sec */
		timeouts_per_sec = 100 / ibpm->timeout;

		if (timeouts_per_sec == 0)
			timeouts_per_sec = 1;
	}

	refresh_time_count = timeouts_per_sec*ibpm->window_size;

	return refresh_time_count;
}

/*
* DESCRIPTION
*	first, send ibpm->window_size amount of records
*	then try to receive MADs, after each MAD
*	received send another MAD until all records
*	were sent. Last, received all MADS that are
*	left.
*	send only records that need to be sent (by sample
*	interval).
*/
static int collect_counters(ibpm_t *ibpm)
{
	record_t *p_rec;
	struct timeval curtime;
	int next_id = 0;
	int i = 0;
	log_t *p_log = &ibpm->log;
	tables_t *table = &ibpm->table[ibpm->active_tbl];
	int ret = 0;
	int index;
	int diff;
        int refresh_time_count = calc_refresh_time_count(ibpm);
        int rcv_mad_count = 0;

	/* init sent records number */
	ibpm->sent = 0;	/* consider renaming */
	ibpm->stats.mad_sent = 0;
	/* init trid array */
	memset(ibpm->trid_arr, 0, ibpm->window_size*sizeof(trid_arr_t));
	gettimeofday(&curtime, NULL);
	if (curtime.tv_sec > ibpm->leasetime) {
		IBPM_LOG(p_log, IBPM_WARN,
				 "lease time expired: current time (%d) > (%d) lease time\n",
				 curtime, ibpm->leasetime);
		return 1;
	}

	init_rec_time(table, &curtime);

	/* send 'window size' amount of records */
	while ((ibpm->sent < ibpm->window_size) && (ibpm->sent < table->rsize) && (i < table->rsize)) {
		p_rec = table->rlinks[i];
		if (p_rec->pc_timeouts > ibpm->max_timeouts_number) {
			/* Don't send record when there were timeouts; assuming port is down */
			i++;
			continue;
		}
		diff = diff_msec(&p_rec->send_time, &curtime);
		if (ibpm->reset_mode != NO_RESET && !p_rec->is_support_extended_pc && diff < -p_ibpm->default_poll_time_msec) {
			IBPM_LOG(p_log, IBPM_ERR, "CRIT: send_time invaild, "
					 "lid=%d port=%d send_time(sec)=%d, curtime(sec)=%d, "
					 "diff(ms)=%d, isample=%d support_extended:%d\n",
					 p_rec->lid, p_rec->port, p_rec->send_time.tv_sec,
					 curtime.tv_sec, diff, p_rec->isample, p_rec->is_support_extended_pc);
			update_rec_time(p_rec, &curtime);
		}
		else if (diff >= 0 || clear_all_phy_cnts) {
			/* check interval time */
			IBPM_LOG(p_log, IBPM_DEBUG,
					 "send MAD: table[%d]: lid=%d port=%d isample=%d "
					 "type=%d devid=0x%x dt=%d\n",
					 p_rec->index, p_rec->lid, p_rec->port,
					 p_rec->isample, p_rec->type, p_rec->devid, diff);
			if (find_free_tr(ibpm, &next_id)) {
				if (send_record(ibpm, p_rec, &curtime, next_id) < 0)
					reset_record(p_rec);
			} else
				IBPM_LOG(p_log, IBPM_ERR, "CRIT: cannot find free trid,"
						 " sent(%d) window(%d)\n", ibpm->sent, ibpm->window_size);
		}
		i++;
	}

        //

	/* receive and send record until all records were sent. */
	while (i < table->rsize) {

		rcv_mad_count++;
		if(refresh_time_count < rcv_mad_count){
			gettimeofday(&curtime, NULL);
			rcv_mad_count = 0;
		}

		/* try to receive mad if there is any */
		if (ibpm->sent) {
			if ((ret = rcv_mad(ibpm, &next_id)) < 0) {
				IBPM_LOG(p_log, IBPM_WARN, "rcv_mad failed. "
						"sent mads %d\n", ibpm->sent);
				continue;
			}
			if (ret > 0) { /* timeout */
				index = ibpm->trid_arr[next_id].index;
				p_rec = &table->rarray[index];
				reset_record(p_rec);
				if (p_rec->type == MAD_TYPE_PC) {
					ibpm->stats.pc_timeouts++;
					p_rec->pc_timeouts++;
					p_rec->status = ret;
					p_rec->to_report = 1;
				} else  if (p_rec->type == MAD_TYPE_PC_EXT) {
					ibpm->stats.pc_timeouts++;
					p_rec->pc_timeouts++;
					p_rec->status = ret;
				} else
					ibpm->stats.vendor_timeouts++;
			}
			/* when sending a reset mad we change the next_id to be -1.
			 * we need to receive another mad in order to have a free slot
			 * to send the next record */
			if (next_id == -1) {
				IBPM_LOG(p_log, IBPM_DEBUG, "sent reset mad, receive another mad\n");
				continue;
			}
		}

		/* search for the next record to send (interval time exceeded) */
		do {
			p_rec = table->rlinks[i++];
			if (p_rec->pc_timeouts > ibpm->max_timeouts_number) {
				continue;
			}
			diff = diff_msec(&p_rec->send_time, &curtime);
			if (ibpm->reset_mode != NO_RESET && !p_rec->is_support_extended_pc && diff < -p_ibpm->default_poll_time_msec) {
				IBPM_LOG(p_log, IBPM_ERR, "CRIT: send_time invaild, "
						 "lid=%d port=%d send_time(sec)=%d, curtime(sec)=%d, "
						 "diff(ms)=%d, isample=%d\n",
						 p_rec->lid, p_rec->port, p_rec->send_time.tv_sec,
						 curtime.tv_sec, diff, p_rec->isample);
			} else if (diff >= 0 || clear_all_phy_cnts) {
				IBPM_LOG(p_log, IBPM_DEBUG,
						 "send MAD: table[%d]: lid=%d port=%d isample=%d "
						 "type=%d devid=0x%x dt=%d\n",
						 p_rec->index, p_rec->lid, p_rec->port,
						 p_rec->isample, p_rec->type, p_rec->devid, diff);
				if (send_record(ibpm, p_rec, &curtime, next_id) < 0)
					reset_record(p_rec);
				break;
			}
		} while (i < table->rsize);
	}

	/* receive all the leftover records */
	while (ibpm->sent > 0) {
		ret = rcv_mad(ibpm, &next_id);
		if (ret > 0) {
			/* request timed out */
			index = ibpm->trid_arr[next_id].index;
			p_rec = &table->rarray[index];
			reset_record(p_rec);
			if (p_rec->type == MAD_TYPE_PC) {
				ibpm->stats.pc_timeouts++;
				p_rec->pc_timeouts++;
				p_rec->status = ret;
				p_rec->to_report = 1;
			} else  if (p_rec->type == MAD_TYPE_PC_EXT) {
				ibpm->stats.pc_timeouts++;
				p_rec->pc_timeouts++;
				p_rec->status = ret;
			}else
				ibpm->stats.vendor_timeouts++;
		} else if (ret < 0) {
			IBPM_LOG(p_log, IBPM_WARN, "rcv_mad failed outstanding mads %d\n",
					ibpm->sent);
		}
	}

	if (ibpm->verbose)
		IBPM_LOG(p_log, IBPM_ERR, "total sent %d / %d\n",
				ibpm->stats.mad_sent, table->rsize);

	return 0;
}

/*
* DESCRIPTION
*	clear all accumulative counters
*/
static void clear_all_counters(ibpm_t *ibpm)
{
	tables_t *table = &ibpm->table[ibpm->active_tbl];
	int new_clear = ibpm->clear_all[ibpm->active_tbl];
	int old_clear = ibpm->clear_all[!ibpm->active_tbl];
	log_t *p_log = &ibpm->log;
	record_t *p_rec = 0;
	sw_rec_t *sw;
	int i;


	if (new_clear != old_clear) {

		clear_all_phy_cnts = 1;/* changed to 1 to reset physical counters */
		struct perf_count tmp;
		memcpy(&tmp, &ibpm->pc_thr, sizeof(tmp));
		memset(&ibpm->pc_thr, 0, sizeof(tmp));

		if(collect_counters(ibpm))
			IBPM_LOG(p_log, IBPM_ERR, "Error collecting port counters\n");
		memcpy(&ibpm->pc_thr, &tmp, sizeof(tmp));
		clear_all_phy_cnts = 0;

		for (i = 0; i < table->rsize; i++) {
			p_rec = table->rlinks[i];
			memset(&p_rec->cur_pc_sum, 0, sizeof(p_rec->cur_pc_sum));
			memset(&p_rec->new_pc_sum, 0, sizeof(p_rec->new_pc_sum));
		}
		for (i = 0; i < table->ssize; i++) {
			sw = &table->sarray[i];
			memset(sw->xmitwait_cur_sum, 0, sizeof(sw->xmitwait_cur_sum));
			memset(sw->xmitwait_new_sum, 0, sizeof(sw->xmitwait_new_sum));
		}
	}
}

/*
* DESCRIPTION
*	add a new vendor record.
*/
static void insert_extended_pc_rec(ibpm_t *ibpm, tables_t *table, record_t *r)
{
	record_t *v = NULL;
	log_t *p_log = &ibpm->log;

	if (table->rsize >= ibpm->table[ibpm->active_tbl].numrecords) {
		IBPM_LOG(p_log, IBPM_ERR, "ERR: max record size %d exceeded\n",
		 ibpm->table[ibpm->active_tbl].numrecords);
		return;
	}

	v = &table->rarray[table->rsize];

	v->port = r->port;
	v->lid  = r->lid;
	v->sw_id = -1;
	v->index = table->rsize;
	v->devid = r->devid;
	v->rev = r->rev;
	v->vendor_xmitwait = 0;
	v->is_support_extended_pc = TRUE;
	v->isample = r->isample;
	v->ireport = r->ireport;
	v->type = MAD_TYPE_PC_EXT;
	v->firstrun = 1;
	table->rlinks[table->rsize] = v;
	table->rsize++;
	r->ext_id = v->index;

	return;
}

/*
* DESCRIPTION
*	add a new switch record.
*/
static void add_switch_rec(tables_t *table, record_t *r)
{
	sw_rec_t *p_sw = &table->sarray[table->ssize];

	/* create new switch record */
	p_sw->lid   = r->lid;
	p_sw->devid = r->devid;
	p_sw->rev   = r->rev >> 8;
	table->ssize++;
}

/*
* DESCRIPTION
*	add a new vendor record.
*/
static int add_vendor_rec(tables_t *table, record_t *r, int rec_type)
{
	record_t *v = &table->rarray[table->rsize];

	v->port = 0;
	v->lid  = r->lid;
	v->sw_id = table->ssize - 1;
	v->index = table->rsize;
	v->devid = r->devid;
	v->rev = r->rev;
	v->vendor_xmitwait = r->vendor_xmitwait;
	v->isample = 1;	/* set default 1 sec interval */
	v->ireport = 1;	/* set default 1 sec interval */
	v->type = rec_type;
	v->firstrun = 1;
	table->rlinks[table->rsize] = v;
	table->rsize++;

	return v->index;
}

/*
* DESCRIPTION
*	insert two new vendor record
*	one for low xmitwait ports and the
*	other for high.
*/
static void insert_vendor_rec(tables_t *table, record_t *r, int *v1_index, int *v2_index)
{
	add_switch_rec(table, r);
	*v1_index = add_vendor_rec(table, r, MAD_TYPE_LOW_XMITWAIT);
	*v2_index = add_vendor_rec(table, r, MAD_TYPE_HI_XMITWAIT);

	/* init record */
	r->sw_id = (r->port < 13) ? *v1_index : *v2_index;
}

static void dump_table(ibpm_t *ibpm)
{
	log_t *p_log = &ibpm->log;
	tables_t *table = &ibpm->table[ibpm->active_tbl];
	record_t *p_rec;
	sw_rec_t *sw;
	int i;

	/* print info */
	if (ibpm->verbose) {
		IBPM_LOG(p_log, IBPM_INFO, "actual expired lease time = %d\n",
				ibpm->leasetime);
		IBPM_LOG(p_log, IBPM_INFO, "total number of sw = %d\n", table->ssize);
		IBPM_LOG(p_log, IBPM_INFO, "total number of rec = %d\n", table->rsize);

		for (i = 0; i < table->rsize; i++) {
			p_rec = table->rlinks[i];
			IBPM_LOG(p_log, IBPM_INFO,
					"table[%d]: lid=%d port=%d "
					"interval=%d devid=0x%x rev=0x%x "
					"type=%d sw_id=%d vend_xmit=%d is_support_extended_pc:%s\n",
					p_rec->index, p_rec->lid, p_rec->port,
					p_rec->ireport, p_rec->devid, p_rec->rev,
					p_rec->type, p_rec->sw_id,
					p_rec->vendor_xmitwait, (p_rec->is_support_extended_pc ? "TRUE" : "FALSE"));
		}
		for (i = 0; i < table->ssize; i++) {
			sw = &table->sarray[i];
			IBPM_LOG(p_log, IBPM_INFO,
					 "sw_table[%d]: lid=%d devid=0x%x rev=0x%x\n",
					 i, sw->lid, sw->devid, sw->rev);
		}
	}
}

/*
* DESCRIPTION
*	parse the config file.
*	skip all lines starting with '#'
*	skip whitespaces and tabs.
*	create vendor records (for switches
*	who dosen't have xmitwait counter in
*	their regular counters).
*/
static int parse_config_file(ibpm_t *ibpm, int enable_retry)
{
	char line[1024];
	char address[16];
	char ufm_protocol[UFM_PROTOCOL_LEN];
	char ufm_url[UFM_URL_LEN];
	int port;
	int checklease = 1;
	int checkaddress = 1;
	int checkport = 1;
	int checkprotocol = 1;
	int checkurl = 1;
	int checkufmkey = 1;
	int checkpolltime = 1;
	int checkresetmode = 1;
	int checkthreshold = 1;
	int checkclearall = 1;
	int checkpsl = (!ibpm->vlr_config.psl_cmdln);
	int checkpeer = (!ibpm->vlr_config.peer_cmdln);
	int parse_retry = 0;
	int rc;
	int increase_port_count = 0;
	struct timeval tv;
	record_t *r;
	log_t *p_log = &ibpm->log;
	config_t *p_config = &ibpm->config;
	tables_t *table;
	int last_lid = 0;
	int v1_index, v2_index;
	int *clear_all;
	char *q, *p;

	ibpm->active_tbl = !ibpm->active_tbl; /* new table id */
	clear_all = &ibpm->clear_all[ibpm->active_tbl];
	table = &ibpm->table[ibpm->active_tbl];
	table->rsize = 0;
	table->ssize = 0;
	v1_index = 0;
	v2_index = 0;

	if (ibpm->table[ibpm->active_tbl].numrecords <
	    ibpm->table[!ibpm->active_tbl].numrecords) {

		ibpm_increase_numrecords(ibpm, 0);
	}

	/* clear old table data */
	memset(table->rarray, 0, table->numrecords*sizeof(record_t));
	memset(table->sarray, 0, MAX_SW_SIZE*sizeof(sw_rec_t));

	p_config->file = fopen(p_config->filename, "r");
	if (!p_config->file) {
		IBPM_LOG(p_log, IBPM_ERR,
				 "ERR: Cannot open config file \'%s\': %m\n",
				 p_config->filename);
		goto Err;
	} else
		IBPM_LOG(p_log, IBPM_INFO, "file %s successfully opened\n",
				 p_config->filename);

	while (fgets(line, sizeof(line) - 1, p_config->file) != NULL) {

		if (table->rsize >= table->numrecords) {

			if (!enable_retry) {
				IBPM_LOG(p_log, IBPM_ERR,"ERR: records number too small."
						 " Cannot complete parsing of the configuration file .\n");
				break;
			}

			while (fgets(line, sizeof(line) - 1, p_config->file) != NULL)
				increase_port_count++;

			ibpm_increase_numrecords(ibpm, increase_port_count+1);
			parse_retry = 1;
			break;
		}
		p = line;

		/* skip rest of the line after '#' */
		q = strchr(p, '#');
		if (q)
			*q = '\0';

		if (checkclearall && (q = strstr(p, "clear_all"))) {
			parse_int(p_log, q, "clear_all", clear_all, 0);
			checkthreshold = 0;
			continue;
		}

		if (checkthreshold && (q = strstr(p, "threshold"))) {
			parse_int(p_log, q, "threshold", &ibpm->threshold, DEFAULT_THRESHOLD);
			checkthreshold = 0;
			continue;
		}

		if (checkresetmode && (q = strstr(p, "reset_mode"))) {
			parse_int(p_log, q, "reset_mode", &ibpm->reset_mode, RESET_EVERY_POLL);
			checkresetmode = 0;
			continue;
		}

		if (checkpolltime && (q = strstr(p, "default_poll_time"))) {
			parse_int(p_log, q, "default_poll_time", &ibpm->default_poll_time, DEFAULT_SAMPLE_INTERVAL);
			ibpm->default_poll_time_msec = ibpm->default_poll_time * 1000;
			checkpolltime = 0;
			continue;
		}

		if (checklease && (q = strstr(p, "leasetime"))) {
			parse_int(p_log, q, "leasetime", &ibpm->leasetime, D_LEASETIME);
			checklease = 0;
			continue;
		}
		if (checkaddress && (q = strstr(p, "address"))) {
			parse_string(p_log, q, "address", address, D_ADDR);
			if (strcmp(address, ibpm->address)) {
				ibpm->update_socket = 1;
				strcpy(ibpm->address, address);
			}

			checkaddress = 0;
			continue;
		}
		if (checkport && (q = strstr(p, "port"))) {
			parse_int(p_log, q, "port", &port, D_PORT);
			if (port != ibpm->port) {
				ibpm->update_socket = 1;
				ibpm->port = port;
			}

			checkport = 0;
			continue;
		}
		if (checkprotocol && (q = strstr(p, "ufm_protocol"))) {
			parse_string(p_log, q, "ufm_protocol", ufm_protocol, UFM_PROTOCOL_UDP);
			if (strcmp(ufm_protocol, ibpm->ufm_protocol)) {
				ibpm->update_socket = 1;
				strcpy(ibpm->ufm_protocol, ufm_protocol);
			}

			checkprotocol = 0;
			continue;
		}
		if (checkurl && (q = strstr(p, "ufm_url"))) {
			parse_string(p_log, q, "ufm_url", ufm_url, "");
			if (strcmp(ufm_url, ibpm->ufm_url)) {
				ibpm->update_socket = 1;
				strcpy(ibpm->ufm_url, ufm_url);
			}

			checkurl = 0;
			continue;
		}
		if (checkufmkey && (q = strstr(p, "ufm_key"))) {
			parse_uint64(p_log, q, "ufm_key", &ibpm->message->header.ufm_key, D_KEY);
			checkufmkey = 0;
			continue;
		}
		if (checkpsl && (q = strstr(p, "path_sl_file"))) {
			parse_file_path(p_log, q, "path_sl_file", &ibpm->vlr_config.psl_file, "");
			checkpsl = 0;
			continue;
		}
		if (checkpeer && (q = strstr(p, "peer_path_file"))) {
			parse_file_path(p_log, q, "peer_path_file", &ibpm->vlr_config.peer_path_file, "");
			checkpeer = 0;
			continue;
		}

		/* skip all whitespaces, tabs or end of lines */
		while (*p == ' ' || *p == '\t' || *p == '\n')
			p++;
		if (*p == '\0')
			continue;

		r = &table->rarray[table->rsize];
		r->index = table->rsize;
		/* the following format is expected:
		  #LID		Port	Interval (sec)		GUID			Device Type ID	Revision
		  1			2		4				0x0008f10400410baf		0x634a		0x000001a1
		*/
		rc = sscanf(p, "%d %d %d %llx %i %i",
					&r->lid, &r->port, &r->ireport,
					(long long unsigned int *)(&r->guid), &r->devid, &r->rev);
		if (rc < 5 || r->port < 1 || r->port > DEV_MAX_NUM_PORTS || r->lid < 1 || r->lid > MAX_UCAST_LID) {
			IBPM_LOG(p_log, IBPM_ERR, "ERR: invalid record format read failed rc=%d\n", rc);
			continue;
		}

		r->firstrun = 1;
		r->type = MAD_TYPE_PC;
		/* during merge set firstrun for new records only */
		table->rlinks[table->rsize] = r;
		table->rsize++;

		r->is_support_extended_pc = is_dev_support_extended_pc(ibpm, r->devid,r->lid);

		if(r->is_support_extended_pc)
		{
			insert_extended_pc_rec(ibpm, table, r);
			r->sw_id = -1;
		}

		r->vendor_xmitwait = check_devid(r->devid);
		if(!r->vendor_xmitwait)
			continue;

		if (r->lid == last_lid)
			r->sw_id = (r->port < 13) ? v1_index : v2_index;
		else {
			last_lid = r->lid;
			if (table->ssize == MAX_SW_SIZE) {
				IBPM_LOG(p_log, IBPM_ERR, "ERR: max switch table size"
						 " %d exceeded\n", MAX_SW_SIZE);
				break;
			}
			insert_vendor_rec(table, r, &v1_index, &v2_index);
		}
	}

	if (fclose(p_config->file))
		IBPM_LOG(p_log, IBPM_ERR, "ERR: %p %s %m, errno: %d\n",
				 p_log->file, p_log->filename, errno);

	if (parse_retry && enable_retry) {
		ibpm->active_tbl = !ibpm->active_tbl; /*get ready for retry*/
		return 1;
	}

	if (checkpsl && ibpm->vlr_config.lid_to_sl)
		freeVlrArray(ibpm, 1);

	if (checkpeer && ibpm->vlr_config.ca_to_sw_lid)
		freeVlrArray(ibpm, 0);

	if (!checkpeer && checkpsl)
		IBPM_LOG(p_log, IBPM_WARN,
				 "Setting peer_path_file without path_sl_file has no effect.\n");

	if (checklease || checkaddress || checkport || checkufmkey)
		goto Err;


	gettimeofday(&tv, NULL);
	/* set leastime to absolute time */
	ibpm->leasetime += tv.tv_sec;
	/* dump table info */
	dump_table(ibpm);

	return 0;

Err:
	/* switch back to old table */
	IBPM_LOG(p_log, IBPM_ERR, "ERR: parsing failed. "
			 "Please check records.conf file.\n"
			 "switching back to old table\n");
	ibpm->active_tbl = !ibpm->active_tbl;
	return -1;
}

/*
* DESCRIPTION
*	compare two records by their
*	lid and port.
*/
int compare(const void *a, const void *b)
{
	int ret = 0;
	record_t *r1 = *(record_t **)a;
	record_t *r2 = *(record_t **)b;

	ret = r1->lid - r2->lid;
	if (ret)
		return ret;

	ret = r1->port - r2->port;
	if (ret)
		return ret;

	ret = r1->type - r2->type;

	return ret;
}

/*
* DESCRIPTION
*	update new records with the old one's
*	data.
*/
static void rec_update(record_t *newrec, record_t *oldrec)
{
	newrec->to_report = oldrec->to_report;
	newrec->cur_time = oldrec->cur_time;
	newrec->last_time = oldrec->last_time;
	newrec->send_time = oldrec->send_time;
	if (newrec->is_support_extended_pc)
		newrec->isample = newrec->ireport;
	else
		newrec->isample = oldrec->isample;
	newrec->firstrun = oldrec->firstrun;
	newrec->cur_pc_sum = oldrec->cur_pc_sum;
	newrec->new_pc_sum = oldrec->new_pc_sum;
	newrec->pcount = oldrec->pcount; /*For counter simulation*/
	/* Zero timeout counters */
	newrec->pc_timeouts = oldrec->pc_timeouts = 0;
}

/*
* DESCRIPTION
*	merge switch records.
*	copy the xmitwait arrays from old
*	switch to the new switch.
*/
static void merge_sw_records(tables_t *newtbl, tables_t *oldtbl)
{
	int newid;
	int oldid;
	sw_rec_t *p_newsw;
	sw_rec_t *p_oldsw;

	for (newid = 0; newid < newtbl->ssize; newid++) {
		p_newsw = &newtbl->sarray[newid];
		for (oldid = 0; oldid < oldtbl->ssize; oldid++) {
			p_oldsw = &oldtbl->sarray[oldid];
			if (p_newsw->lid != p_oldsw->lid)
				continue;
			/* copy old switch cumulative counters to new switch */
			memcpy(p_newsw->xmitwait_cur_sum,
				p_oldsw->xmitwait_cur_sum,
				sizeof(p_newsw->xmitwait_cur_sum));
			memcpy(p_newsw->xmitwait_new_sum,
				p_oldsw->xmitwait_new_sum,
				sizeof(p_newsw->xmitwait_new_sum));
			break;
		}
	}

}

/*
* DESCRIPTION
*	merge records from old and new tables.
*	if a record in the new tables
*	exists in the old table - copy
*	it's data.
*/
static void merge_records(tables_t *newtbl, tables_t *oldtbl)
{
	int newid = 0;
	int oldid = 0;
	record_t *newrec;
	record_t *oldrec;

	while (newid < newtbl->rsize && oldid < oldtbl->rsize) {
		newrec = newtbl->rlinks[newid];
		oldrec = oldtbl->rlinks[oldid];

		/* new record is bigger then old record */
		if (newrec->lid > oldrec->lid) {
			oldid++;
			continue;
		}

		/* new record is smaller then old record */
		if (newrec->lid < oldrec->lid) {
			newid++;
			continue;
		}

		/* lids match - check ports match */
		if (newrec->port > oldrec->port) {
			oldid++;
			continue;
		}

		if (newrec->port < oldrec->port) {
			newid++;
			continue;
		}

		/* lids and port match - check type match */
		if (newrec->type > oldrec->type) {
			oldid++;
			continue;
		}

		if (newrec->type < oldrec->type) {
			newid++;
			continue;
		}

		/* same (old) record exists in the new table */
		/* copy old records data into the new record */
		rec_update(newrec, oldrec);
		newid++;
		oldid++;
	}
}

/*
* DESCRIPTION
*	merge the new and old tables.
*/
static void merge_tables(tables_t *newtbl, tables_t *oldtbl)
{
	merge_records(newtbl, oldtbl);
	merge_sw_records(newtbl, oldtbl);
}

/*
* DESCRIPTION
*	check if port is active.
*   if retries == 0 run with read_baselid == 1 to update ibpm->baselid (for vlr)
*	return 0 - if port is active.
*	return 1 - if port is not active or failed to read port.
*/
static int check_port(ibpm_t *ibpm, int read_baselid)
{
	uint8_t data[IB_SMP_DATA_SIZE];
	char buf[1024];
	ib_portid_t self = { 0 };
	int retries;
	int total_retries;
	int state;
	int sm_state;
	int phystate;
	int baselid;

	for(retries = 1, total_retries = 1 ;
		read_baselid || retries <= ibpm->port_poll_retries;
		retries++, total_retries++) {
		if (!smp_query_via(data, &self, IB_ATTR_PORT_INFO, ibpm->ca_port, 0, ibpm->srcport)) {
			IBPM_LOG(&ibpm->log, IBPM_ERR, "Failed to read port info (retry %d/%d)\n",
					retries, ibpm->port_poll_retries);

			if (read_baselid)
				return -1;

			sleep(ibpm->port_poll_interval);
			continue;
		}
		mad_decode_field(data, IB_PORT_STATE_F, &state);
		mad_decode_field(data, IB_PORT_PHYS_STATE_F, &phystate);
		mad_decode_field(data, IB_PORT_LID_F, &baselid);


		IBPM_LOG(&ibpm->log, IBPM_DEBUG, "checking ca %s port %d state %d phystate %d baselid :%d\n",
				ibpm->ca, ibpm->ca_port, state, phystate, baselid);

		if (state == 4) {
			if (baselid != ibpm->baselid) {
				ibpm->baselid = (uint16_t)baselid;
				setRereadVlrFiles(ibpm);
			}
			return 0;
		}

		if (read_baselid)
			return -1;


		/* port is not active continue */
		if(state != 1) {/* Not Down */
			if (!smp_query_via(data, &self, IB_ATTR_SMINFO, ibpm->ca_port, 0, ibpm->srcport)) {
				snprintf(buf, sizeof(buf),
					"port state: %s(%u) / %s(%u) , SM is unreachable, retry (%d/%d)\n",
					get_state_str(state), state, get_phy_state_str(phystate), phystate,
					retries, ibpm->port_poll_retries);
			}
			else {
				mad_decode_field(data, IB_SMINFO_STATE_F, &sm_state);

				snprintf(buf, sizeof(buf),
					"port state: %s(%u) / %s(%u) SM state is %s (%u). retry (#%u) until the port will change state \n",
					get_state_str(state), state, get_phy_state_str(phystate), phystate,
					SM_STATE_STR(sm_state), sm_state, total_retries);

				retries = 0;
			}

		}
		else {
			snprintf(buf, sizeof(buf),
				"port state: %s(%u) / %s(%u) retry (%d/%d)\n",
				get_state_str(state), state, get_phy_state_str(phystate), phystate,
				retries, ibpm->port_poll_retries);
		}

		IBPM_LOG(&ibpm->log, IBPM_ERR, "%s", buf);
		syslog(LOG_ALERT, "%s", buf);
		sleep(ibpm->port_poll_interval);
	}

	if (retries >= ibpm->port_poll_retries) {
		snprintf(buf, sizeof(buf),
			"shutting down IBPM - port state is not Active after %d sec\n",
			ibpm->port_poll_interval * ibpm->port_poll_retries);

		IBPM_LOG(&ibpm->log, IBPM_ERR, "CRIT %s", buf);
		syslog(LOG_ALERT, "%s", buf);
		ibpm_exit(ibpm);
		exit(1);
	}
	return -1;
}

static int is_modified(ibpm_t *ibpm,
					   const char * filename,
					   time_t *mtime,
					   int *modified)
{
	struct stat file_stat;
	log_t *p_log = &ibpm->log;
	*modified=0;

	if (stat(filename, &file_stat) < 0) {
		IBPM_LOG(p_log, IBPM_ERR, "ERR: stat failed filename %s %m\n", filename);
		return -1;
	}

	/* check file's modify time*/
	if (*mtime == file_stat.st_mtime) {
		IBPM_LOG(p_log, IBPM_DEBUG,
				 "file:%s was not modified\n", filename);
		return 0;
	}

	*modified=1;
	*mtime = file_stat.st_mtime;

	return 0;
}

static int check_vlr_config(ibpm_t *ibpm)
{
	/* consider not using static variable */
	int is_conf_modified;

	if (ibpm->vlr_config.peer_path_file) {

		if (is_modified(ibpm,
						ibpm->vlr_config.peer_path_file,
						&ibpm->vlr_config.peer_path_mtime,
						&is_conf_modified))
			return -1;

		if (is_conf_modified) {
			ibpm->vlr_config.peer_rc = parsePeerPathFile(ibpm);
			ibpm->vlr_config.psl_mtime = 0;
		}

		if (ibpm->vlr_config.peer_rc)
			return -1;
	}

	//PSL
	if (ibpm->vlr_config.psl_file) {

		if (is_modified(ibpm,
						ibpm->vlr_config.psl_file,
						&ibpm->vlr_config.psl_mtime,
						&is_conf_modified))
			return -1;

		if (is_conf_modified)
			ibpm->vlr_config.psl_rc = parsePSLFile(ibpm);
		if (ibpm->vlr_config.psl_rc)
			return -1;
	}

	return 0;
}

/*
* DESCRIPTION
*	check if config file has been change.
*	if so, parse it and merge it with the
*	old table.
*/
static void check_config(ibpm_t *ibpm)
{
	/* consider not using static variable */
	//log_t *p_log = &ibpm->log;
	tables_t *table;
	tables_t *oldtable;
	int parse_config_rc = 0;
	int is_conf_modified;

	if(is_modified(ibpm,
				   ibpm->config.filename,
				   &ibpm->config.mtime,
				   &is_conf_modified))
		return;

	if(!is_conf_modified)
		return;

	parse_config_rc = parse_config_file(ibpm, 1);
	if (parse_config_rc == 1)
		parse_config_rc = parse_config_file(ibpm, 0); /* retry after realloc */
	if (parse_config_rc != 0)
		return;

	if (ibpm->reset_mode == RESET_EVERY_POLL)
		ibpm->threshold = 100;

	init_threshold(ibpm);

	ibpm->stats.switch_tbl++;
	table = &ibpm->table[ibpm->active_tbl];
	oldtable = &ibpm->table[!ibpm->active_tbl];

	/* sort the new table */
	qsort(table->rlinks, table->rsize, sizeof(record_t *), compare);

	/* merge the old and new tables */
	merge_tables(table, oldtable);

	clear_all_counters(ibpm);

	/* socket initialization */
	init_socket(ibpm);

	/* check switch revision */
	check_sw_rev(ibpm, table);
}

static void dump_conf_template(void)
{
	printf("\n#IBPM conf file should be located at: %s\n\n",
		   DEFAULT_CONF_FILENAME);
	printf("debug   2            # set debug level 2 [0-4]\n"
		   "verbose 0            # disable/enable verbosity [0/1]\n"
		   "csv     0            # disable/enable csv [0/1]\n"
		   "report  0            # disable/enable report level 0 [0/1]\n"
		   "poll_retries  6      # polling retries when port is not ACTIVE\n"
		   "poll_timeout  5      # polling timeout per retry when port is not ACTIVE\n\n" );
}

static int get_options(int argc, char *argv[], ibpm_t *ibpm)
{
	uint32_t next_option;
	const char *const short_option = "c:d:n:l:t:C:P:w:E:R:T:m:s:p:Vhfxzve";
	log_t *p_log = &ibpm->log;
	config_t *p_config = &ibpm->config;
	char *max_timeouts = NULL;

	/*
		In the array below, the 2nd parameter specifies the number
		of arguments as follows:
		0: no arguments
		1: argument
		2: optional
	*/
	const struct option long_option[] = {
		{"version", 0, NULL, 'V'},
		{"verbose", 0, NULL, 'v'},
		{"csv", 0, NULL, 'x'},
		{"logfile", 1, NULL, 'l'},
		{"Ca", 1, NULL, 'C'},
		{"Port", 1, NULL, 'P'},
		{"foreground", 0, NULL, 'f'},
		{"conf", 1, NULL, 'c'},
		{"debug", 1, NULL, 'd'},
		{"num", 1, NULL, 'n'},
		{"window", 1, NULL, 'w'},
		{"timeout", 1, NULL, 't'},
		{"template", 0, NULL, 'z'},
		{"eraselog", 0, NULL, 'e'},
		{"errrate", 0, NULL, 'E'},
		{"pollretries", 1, NULL, 'R'},
		{"pollinterval", 1, NULL, 'T'},
		{"help", 0, NULL, 'h'},
		{"max_timeouts", 1, NULL, 'm'},
		{"path_sl_file", 1, NULL, 's'},
		{"peer_path_file", 1, NULL, 'p'},
		{NULL, 0, NULL, 0}	/* Required at the end of the array */
	};
	/* first get option from the environment */
	max_timeouts = getenv("MAX_TIMEOUTS");
	if (max_timeouts)
		ibpm->max_timeouts_number = (unsigned int)atoi(max_timeouts);

	do {
		next_option = getopt_long_only(argc, argv, short_option,
									   long_option, NULL);
		switch (next_option) {
		case 'c':
			p_config->filename = strdup(optarg);

			break;
		case 'w':
			ibpm->window_size = strtol(optarg, NULL, 0);
			break;

		case 'C':
			ibpm->ca = strdup(optarg);
			break;

		case 'P':
			ibpm->ca_port = strtoul(optarg, 0, 0);
			break;

		case 'd':
			p_log->dbg_lvl = strtol(optarg, NULL, 0);
			break;

		case 'V':
			printf("version: %s\n", IBPM_VERSION);
			return -1;
			break;

		case 'v':
			ibpm->verbose = 1;
			break;

		case 'T':
			ibpm->port_poll_interval = strtoul(optarg, 0, 0);
			break;

		case 'R':
			ibpm->port_poll_retries = strtoul(optarg, 0, 0);
			break;

		case 'n':
			ibpm->numrecords = strtol(optarg, NULL, 0);
			break;

		case 'f':
			p_log->file = NULL;
			p_log->filename = NULL;
			printf("ibpm is running in foreground\n");
			break;

		case 'l':
			p_log->filename = strdup(optarg);
			break;

		case 't':
			ibpm->timeout = strtol(optarg, NULL, 0);
			break;

		case 'x':
			ibpm->csv = 1;
			break;

		case 'z':
			dump_conf_template();
			return -1;
			break;

		case 'e':
			ibpm->eraselog = 1;
			break;

		case 'E':
			ibpm->errrate = strtol(optarg, NULL, 0);
			break;

		case 'h':
		case '?':
		case ':':
			show_usage();
			return -1;
			break;

		case 'm':
			ibpm->max_timeouts_number = (unsigned int)strtoul(optarg, 0, 0);
			break;

		case 's':
			ibpm->vlr_config.psl_file = strdup(optarg);
			ibpm->vlr_config.psl_cmdln = 1;
			break;

		case 'p':
			ibpm->vlr_config.peer_path_file = strdup(optarg);
			ibpm->vlr_config.peer_cmdln = 1;
			break;

		case -1:
			break;	/* done with option */
		default:	/* something wrong  */
			abort();
		}
	} while (next_option != -1);

	return 0;
}

int main(int argc, char *argv[])
{
	ibpm_t ibpm;
	struct timeval tstart;
	struct timeval tmid;
	struct timeval tend;
	int sleep_time;
	int sleep_usec;
	log_t *p_log = &ibpm.log;
	tables_t *table;
	int tcollect;
	int vlr_config_rc=0;
	clear_all_phy_cnts = 0;
	p_ibpm = &ibpm;
	memset(p_ibpm, 0, sizeof(ibpm_t));

	/* set defaults */
	ibpm_default_init(&ibpm);

	/* get options from user */
	if (get_options(argc, argv, &ibpm) < 0) {
		IBPM_LOG(p_log, IBPM_WARN, "exiting...\n");
		return 1;
	}

	/* ibpm init */
	ibpm_init(&ibpm);

	while (1) {
		check_config(&ibpm);

		if (ibpm.port_poll_retries) {
			check_port(&ibpm, 0);
		} else if (ibpm.vlr_config.psl_file && check_port(&ibpm, 1)){
			IBPM_LOG(p_log, IBPM_ERR, "Failed to read baselid.\n");
		}

		vlr_config_rc = check_vlr_config(&ibpm);

		/* set table to active table */
		table = &ibpm.table[ibpm.active_tbl];
		gettimeofday(&tstart, NULL);
		if (table->rsize != 0 && !vlr_config_rc && collect_counters(&ibpm) == 0) {
			gettimeofday(&tmid, NULL);
			tcollect = diff_usec(&tstart, &tmid);
			if (ibpm.stats.mad_sent) {
				ibpm.stats.avg_send_time = tcollect / ibpm.stats.mad_sent;
				if (ibpm.stats.max_avg < ibpm.stats.avg_send_time)
					ibpm.stats.max_avg = ibpm.stats.avg_send_time;
				if (ibpm.stats.min_avg > ibpm.stats.avg_send_time)
					ibpm.stats.min_avg = ibpm.stats.avg_send_time;
			}

			ibpm_report(&ibpm);
			gettimeofday(&tend, NULL);
			if (ibpm.verbose && table->rsize) {
				printf("tcollect = %d ms\n", tcollect/1000);
				printf("treport  = %d ms\n",
					   diff_usec(&tstart, &tend)/1000);
			}
		}
		gettimeofday(&tend, NULL);

		if (ibpm.print_stats)
			dump_statistics(&ibpm);

		if (ibpm.reset_stats)
			reset_statistics(&ibpm);

		if (ibpm.read_conf)
			read_configuration(&ibpm);

		/* sleep to complete 1 sec */
		sleep_time = diff_usec(&tstart, &tend);
		if (sleep_time < 1000000) {
			sleep_usec = 1000000 - sleep_time;
			IBPM_LOG(p_log, IBPM_DEBUG,
					 "sleeping for %d usec\n", sleep_usec);
			usleep(sleep_usec);
		} else
			IBPM_LOG(p_log, IBPM_DEBUG,
					"%d usec between sendtime to endtime\n", sleep_time);
	}
}
